'''
A lunarcord extension that aims to make the usage of `Components`, `Buttons` and `SelectMenus` easier with the use of `View` objects.
Below is a small snippet showcasing how this extension can possibly be used to create a single `Button` and send it:

```
from lunarcord import ui, Bot

bot = Bot()

@bot.event()
async def onReady():

    view = ui.View()
    
    @view.button("Click Me!", style=ui.ButtonStyle.Green)
    async def button1(interaction):
    
        await interaction.send("Click!", ephemeral=True)
        
    channel = await bot.fetchChannel(12345678)
    await channel.send(f"Logged in as **{bot}**!", view=view)
    
bot.run("EXAMPLE123TOKEN45")
```

The above example would create a `Bot` that, upon logging in, sends a new `Message` in the `Channel` of ID `12345678` which contains a button with the label "Click Me!". Upon clicking the button, an ephemeral `Message` will appear, reading "Click!".
For more examples, read more about `Views` in lunarcord's official docs (`SOON`).
'''

from typing import TypeVar, Generic, Callable
from .__core__.requests import asyncio
from .basetypes import Channel, Context, Interaction, Channel, Message, Emoji
from .__core__.addons import Utils
from .__core__.types import Registrable
from .errors import NotRegisteredError

import time

T = TypeVar('T')

class Component:
    
    def __init__(self, type: int, customID: str = None, row: int = None, callbacks: list = []):
        
        '''
        A base class for all `Component` types (`ActionRow`, `Button`, `SelectMenu`, `Modal`).
        '''
        
        if not isinstance(type, int):
            
            try:
                type = int(type)
                
            except:
                type = 1
        
        self.type: int = type
        
        if customID:
            self.customID: str = customID
        
        if row:
            self.row: int = row
            
        self.callbacks = callbacks
        self.bot = None
        self.message: Message = None
        self.underlying: View = None
        
    @property
    def size(self):
        
        '''
        How much size of an `ActionRow` this `Component` takes. This is usually either `1` or `5`.
        '''
        
        if self.type == View.TEXT_INPUT:
            
            return 5
        
        if self.type == View.BUTTON:
            
            return 1
        
        return 0
    
    @property
    def attached(self):
        
        '''
        Whether this `Component` is part of a `View`.
        '''
        
        return self.underlying is not None
    
    @property
    def view(self):
        
        '''
        The `View` this `Component` is attached to. If it isn't attached to any views, creates a new one containing it and returns it.
        '''
        
        if not self.attached:
            
            return self.toView()
        
        return self.underlying
    
    def toView(self):
        
        '''
        Creates a new `View` which consists of this `Component` only.
        '''
        
        return View(self)
    
    def addTo(self, view):
        
        '''
        Adds this `Component` to the given `View`. Returns the `ActionRow` if any.
        '''
        
        if view and type(view) is View:
            
            return view.addComponent(self)
        
        return None
    
    def remove(self):
        
        '''
        Removes this `Component` from its underlying `View`, if any, and returns the `Row` it was in.
        '''
        
        if self.attached:
            
            return self.underlying.removeComponent(self)
        
        return None
    
    def setView(self, view):
        
        '''
        Removes this `Component` from its current `View`, if any, and adds it to the new one.
        '''
        
        self.remove()
        return self.addTo(view)
        
    def toJson(self) -> dict:
        
        '''
        Converts this `Component` to a JSON-like `dict`.
        '''
        
        jsonified = {
            'type': self.type
        }
        
        if self.type != View.ACTION_ROW and self.customID is not None:
            
            jsonified['custom_id'] = self.customID
            
        return jsonified
        
class ActionRow(Generic[T], Component):
    
    def __init__(self, idx: int):
        
        '''
        Represents an action row component with other sub-components in it.
        '''
        
        super().__init__(View.ACTION_ROW)
        
        self.components: list[Component] = []
        self.totalSpace: int = 5
        self.idx: int = idx
        
    @property
    def size(self):
        
        '''
        Returns the total occupied `ActionRow` size for this action row based on all components inside it.
        '''
        
        total = 0
        
        for component in self.components:
            
            total += component.size
            
        return total
            
    @property
    def length(self):
        
        '''
        The amount of `Component` objects inside this `ActionRow` action row.
        '''
        
        return len(self.components)
            
    @property
    def space(self):
        
        '''
        Returns how much component size / space is left in this `ActionRow` action row.
        '''
        
        return self.totalSpace - self.size
    
    @property
    def hasSpaceForButtons(self) -> bool:
        
        '''
        Whether you can add more buttons to this `ActionRow`, or not if it's already full.
        '''
        
        return self.space >= 1
    
    @property
    def hasSpaceForSelect(self) -> bool:
        
        '''
        Whether you can add a `SelectMenu` to this `ActionRow`, or not if there are other components already.
        '''
        
        return self.space >= 5
        
    def hasSpaceFor(self, component: Component):
        
        '''
        Returns a `bool` representing whether you can and should add this `Component` inside this `ActionRow` or not.
        '''
        
        if type(component) is Button:
            
            return self.hasSpaceForButtons
        
        elif type(component) is SelectMenu:
            
            return self.hasSpaceForSelect
        
        return False
    
    def extend(self, *rows):
        
        '''
        Extends this `ActionRow` action row by adding as many components as possible from the other `rows` to it.
        '''
        
        for row in rows:
            
            if type(row) is not ActionRow:
                
                continue
            
            for component in row:
                
                if self.hasSpaceFor(component):
                    
                    self.addComponent(component)
        
    def toJson(self):
        
        components = []
        type = self.type
        
        for component in self.components:
            
            components.append(
                component.toJson()
            )
            
        jsonified = {
            'type': self.type,
            'components': components
        }
        
        return jsonified
    
    def addComponent(self, component: Component):
        
        '''
        Adds a sub-component to this `ActionRow` component.
        '''
        
        if not self.hasSpaceFor(component):
            raise IndexError('Not enough action row space')
        
        component.row = self.idx
        self.components.append(component)
        
    def removeComponent(self, component: Component):
        
        '''
        Removes an already added component from this `ActionRow` action row. Raises `ValueError` if this `Component` is not present.
        '''
        
        try:
            
            self.components.remove(component)
            
        except ValueError:
            
            raise ValueError('Component not in action row')
        
    def __iter__(self):
        
        return self.components.__iter__()
        
class ButtonStyle:
    
    class Primary: id = 1
    class Secondary: id = 2
    class Success: id = 3
    class Danger: id = 4
    class Link: id = 5
    
    class Blurple: id = 1
    class Gray: id = 2
    class Grey: id = 2
    class Green: id = 3
    class Red: id = 4
        
class Button(Component):
    

    
    def __init__(self, name: str = None, customID: str = None, style: ButtonStyle = ButtonStyle.Primary, emoji: str | int = None, row: int = None, callback: Callable = None):
        
        '''
        A clickable message `Component` that can have one or more callbacks.
        
        Parameters
        ----------
        name: `str`
            The button's name/title/label. This will show up on Discord.
            
        customID: `str`
            A custom ID to be used for identifying this button. Must be unique among all other components in its `View`. If not given, it is automatically derived from the button's label for you.
            
        style: `ButtonStyle`
            A ButtonStyle, or a string representing a style, or an int representing a style's ID.
            
        row: `int`
            The row this should be placed in in your `View`. If not given, Lunarcord automatically handles it for you.
            
        callback: `Callable`
            A callback function that should be ran when someone on Discord presses this button.
        '''
        
        if name is None:
            name = 'Button'
            
        if customID is None:
            customID = name.lower().replace(' ', '_')
            
        try:
            style = style.id
            
        except:
            
            try:
                style = int(style)
                
            except:
                
                if type(style) is str:

                    style = style.lower()
                    
                    match style:
                        case 'primary', 'blurple':
                            style = 1
                        case 'secondary', 'gray', 'grey':
                            style = 2
                        case 'success', 'green':
                            style = 3
                        case 'danger', 'red':
                            style = 4
                        case 'link', 'url':
                            style = 5
                    
                else:
                    style = 1
            
        super().__init__(View.BUTTON, customID)
            
        self.name: str = name
        self.row: int = row
        self.style: int = style
        self.callbacks: list = []

        self.emoji: Emoji | None = None

        if type(emoji) in (str, int):
            emoji = Emoji.create(emoji)

        if type(emoji) is Emoji:
            self.emoji = emoji
        
        if callback is not None:
            self.bind(callback)
            
    def connect(self, callback):
        
        '''
        Add an on clicked callback for this `Button`. When someone on Discord clicks this `Button`, the connected `callback` will be called.
        '''
        
        if callable(callback):
            
            self.callbacks.append(callback)
            
    def bind(self, *callbacks):
        
        '''
        Similarly to `Button.connect()`, this allows you to add another callback. The only difference is that you can pass multiple callbacks at once here.
        '''
        
        for callback in callbacks:
            
            self.connect(callback)
            
    def callback(self, func = None):
        
        '''
        You can use this as a decorator for your callback instead of using `Button.connect`.
        '''
        
        if func is not None and callable(func):
            
            return self.connect(func)
        
        return self.connect
        
    def toJson(self):
        
        jsonified = {
            'type': self.type,
            'label': self.name,
            'custom_id': self.customID,
            'style': self.style
        }

        if self.emoji is not None:
            
            emoji = {
                'name': self.emoji.name,
                'id': self.emoji.id,
                'animated': self.emoji.animated
            }

            jsonified['emoji'] = emoji
        
        return jsonified
    
    def __str__(self):
        
        return f'<Button name="{self.name}" customID="{self.customID}" row={self.row}>'
    
    async def __call__(self, interaction, *args, **kwargs):
        
        '''
        Execute all bound/connected callback functions as if the button has been clicked.
        '''
        
        from .__core__.addons import Utils
        
        coros = []
        
        for callback in self.callbacks:
            
            coros.append(
                
                Utils.execute(callback, interaction, *args, self.bot, **kwargs)
                
            )
            
        await asyncio.gather(*coros)
    
class SelectMenuOption:
    
    def __init__(
        self,
        name: str,
        value: str = None,
        description: str = None,
        emoji: str | int = None,
        callback: Callable = None
    ):
        
        if value is None:
            
            value = name.lower().replace(' ', '_')
        
        self.name: str = name
        self.value: str = value
        self.description: str = description
        self.callbacks: list[Callable] = []

        self.emoji: Emoji | None = None

        if type(emoji) in (str, int):
            emoji = Emoji.create(emoji)

        if type(emoji) is Emoji:
            self.emoji = emoji
        
        if callback is not None:
            self.connect(callback)

        self.bot = None
        self.message: Message = None
            
    def connect(self, callback: Callable):
        
        '''
        Connect a `Callback` function to be executed when this `SelectMenuOption` is called.
        '''
        
        if callable(callback):
            
            self.callbacks.append(callback)
            
    def bind(self, *callbacks: Callable):
        
        '''
        Similar to `SelectMenuOption.connect()`, but allows you to connect multiple `callback` functions by passing all of them at once.
        '''
        
        for callback in callbacks:
            
            self.connect(callback)
            
    def callback(self, func = None):
        
        '''
        You can use this as a decorator for your callback instead of using `SelectMenuOption.connect` or `SelectMenuOption.bind`.
        '''
        
        if func is not None and callable(func):
            
            self.connect(func)
        
    def copy(self):
        
        '''
        Returns a shallow copy of this `SelectMenuOption` option.
        '''
        
        return SelectMenuOption(self.name, self.value, self.description, self.emoji)
        
    def toJson(self):
        
        data = {
            'label': self.name,
            'value': self.value
        }
        
        if self.description is not None:
            
            data['description'] = self.description
            
        if self.emoji is not None:
            
            emoji = {
                'name': self.emoji.name,
                'id': self.emoji.id,
                'animated': self.emoji.animated
            }
            
            data['emoji'] = emoji
            
        return data
    
    def __str__(self):
        
        string = f'<SelectMenuOption name="{self.name}"'
            
        if self.emoji is not None:
            
            string += f' emoji={self.emoji}'
            
        return string + '>'
    
    def __repr__(self):
        
        return self.name
    
    async def __call__(self, interaction, *args, **kwargs):
        
        '''
        Execute all bound/connected callback functions as if the option has been clicked.
        '''
        
        from .__core__.addons import Utils
        
        coros = []
        
        for callback in self.callbacks:
            
            coros.append(
                
                Utils.execute(callback, interaction, *args, self.bot, **kwargs)
                
            )
            
        await asyncio.gather(*coros)
    
    
class SelectMenu(Component):
    
    def __init__(self, name: str = None, options: list[SelectMenuOption] = None, customID: str = None, row: int = None, minvalues: int = 1, maxvalues: int = 1, callback: Callable = None):
        
            
        if type(options) not in (list, tuple, set):
            
            options = []
            
        options = [option for option in options if type(option) is SelectMenuOption]
        
        if customID is None:
            
            value = name if name is not None else '_'.join([option.value.replace(' ', '_') for option in options]) if len(options) >= 1 else 'Select Menu'
            customID = value.lower().replace(' ', '_')
            
        super().__init__(View.STRING_SELECT, customID)
        
        self.name: str = name
        self.callbacks: list[Callable] = []
        self.row: int = row
        self.options: list[SelectMenuOption] = options
        self.min: int = minvalues
        self.max: int = maxvalues
        
        if callback:
            
            self.bind(callback)
            
    def connect(self, callback):
        
        '''
        Add an on clicked callback for this `SelectMenu`. When someone on Discord clicks this `Selectmenu`, the connected `callback` will be called.
        '''
        
        if callable(callback):
            
            self.callbacks.append(callback)
            
    def bind(self, *callbacks):
        
        '''
        Similarly to `SelectMenu.connect()`, this allows you to add another callback. The only difference is that you can pass multiple callbacks at once here.
        '''
        
        for callback in callbacks:
            
            self.connect(callback)
            
    def callback(self, func = None):
        
        '''
        You can use this as a decorator for your callback instead of using `SelectMenu.connect` or `SelectMenu.bind`.
        '''
        
        if func is not None and callable(func):
            
            self.connect(func)
    
    def addSelectOption(self, option: SelectMenuOption):
        
        '''
        Adds a `SelectMenuOption` to the select menu's options.
        '''
        
        if type(option) is SelectMenuOption:
            self.options.append(option)
    
    def addOption(
        self,
        name: str,
        value: str = None,
        description: str = None,
        emoji: str | int = None
    ):
        
        '''
        Creates a new `SelectMenuOption` and adds it to the `SelectMenu`. If you want to add a `SelectMenuOption` object that you have, use `SelectMenu.addSelectOption()`.
        '''
        
        newOption = SelectMenuOption(name, value, description, emoji)
        self.addSelectOption(newOption)
        return newOption
    
    def option(self, name: str = None, value: str = None, description: str = None, emoji: str | int = None):
        
        '''
        A decorator that creates a new `SelectMenuOption` with the decorated function as its `Callback`, adds it to the `SelectMenu`, and returns it.
        '''
        
        def decorator(decorated: Callable) -> Button:
            
            nonlocal self, name, value, description, emoji
        
            option = self.addOption(
                
                name = name,
                value = value,
                description = description,
                emoji = emoji
                
            )
            
            option.bind(decorated)
            return option
            
        return decorator
    
    
    def toJson(self):

        if self.min < 1:
            self.min = 1

        if self.max > len(self.options):
            self.max = -1

        if self.max == -1:
            self.max = len(self.options)
        
        data = {
            'type': self.type,
            'custom_id': self.customID,
            'min_values': self.min,
            'max_values': self.max,
            'options': [option.toJson() for option in self.options]
        }
        
        if self.name:
            
            data['placeholder'] = self.name
            
        return data
    
    async def __call__(self, interaction, *args, **kwargs):
        
        '''
        Execute all bound/connected callback functions as if an option of the select menu has been clicked.
        '''
        
        from .__core__.addons import Utils
        
        coros = []
        
        for callback in self.callbacks:
            
            coros.append(
                
                Utils.execute(callback, interaction, *args, self.bot, **kwargs)
                
            )
            
        await asyncio.gather(*coros)
    
    def __str__(self):
        
        return f'<SelectMenu name="{self.name}" customID="{self.customID}" row={self.row}>'
    
class TextInputStyle:
    
    class Short: '''Single-line text input.'''; type = 1
    class Paragraph: '''Multi-line text input.'''; type = 2

    class Small: '''Multi-line text input.'''; type = 1
    class Big: '''Multi-line text input.'''; type = 2
    
    class SingleLine: '''Multi-line text input.'''; type = 1
    class MultiLine: '''Multi-line text input.'''; type = 2
    
    STYLES: list[type] = [Short, Paragraph, Small, Big, SingleLine, MultiLine]
    
class TextInput:
    
    def __init__(
        self,
        name: str = None,
        style: TextInputStyle = TextInputStyle.Short,
        minLength: int = None,
        maxLength: int = None,
        placeholder: str = None,
        required: bool = True,
        customID: str = None,
        callback: Callable = None
    ):
        
        '''
        A text-input object for `Modal` Components.
        '''
        
        if minLength is None or type(minLength) is not int:
            
            try: minLength = int(minLength)
            except: minLength = 1
            
        if maxLength is None or type(maxLength) is not int:
            
            try: maxLength = int(maxLength)
            except: maxLength = 4000
            
        if customID is None:
            
            customID = name.lower().replace(' ', '_')
            
        if type(style) is not int and style not in TextInputStyle.STYLES:
            
            style = TextInputStyle.Short
            
        else:
            
            if type(style) is not int:
                
                style = style.type
        
        self.bot = None
        self.name: str = name
        self.style: TextInputStyle = style
        self.customID: str = customID
        self.minLength: int = minLength
        self.maxLength: int = maxLength
        self.placeholder: str = placeholder
        self.required: bool = required
        self.value: str = None
        self.callbacks: list[Callable] = []
        
        if callback and callable(callback):
            
            ...
            
    def connect(self, callback: Callable):
        
        '''
        Connect a `Callback` function to be executed when this `TextInput` is filled.
        '''
        
        if callable(callback):
            
            self.callbacks.append(callback)
            
    def bind(self, *callbacks: Callable):
        
        '''
        Similar to `TextInput.connect()`, but allows you to connect multiple `callback` functions by passing all of them at once.
        '''
        
        for callback in callbacks:
            
            self.connect(callback)
            
    def callback(self, func):
        
        '''
        You can use this as a decorator for your callback instead of using `TextInput.connect` or `TextInput.bind`.
        '''
        
        if func is not None and callable(func):
            
            self.connect(func)
            
    def toJson(self) -> dict[str, int | list[dict[str]]]:
        
        row = {
            'type': View.ACTION_ROW,
            'components': []
        }
        
        data = {
            'type': View.TEXT_INPUT,
            'custom_id': self.customID,
            'label': self.name,
            'style': self.style,
            'min_length': self.minLength,
            'max_length': self.maxLength,
            'required': self.required
        }
        
        if self.placeholder is not None:
            data['placeholder'] = self.placeholder
        
        row['components'].append(data)
        
        return row
    
    def __str__(self):
        
        return f'<TextInput name="{self.name}" customID="{self.customID}" value="{self.value}">'
    
    def __repr__(self):
        
        return self.name.__repr__()
    
    async def __call__(self, interaction, *args, **kwargs):
        
        '''
        Execute all bound/connected callback functions as if this input has been filled.
        '''
        
        from .__core__.addons import Utils
        
        coros = []
        
        for callback in self.callbacks:
            
            coros.append(
                
                Utils.execute(callback, interaction, *args, self.bot, **kwargs)
                
            )
            
        await asyncio.gather(*coros)
    
    
        
        
    
class Modal():
    
    def __init__(self, name: str = None, inputs: list[TextInput] = None, customID: str = None, callback: Callable = None):
        
        '''
        A message `Component` made up of multiple `TextInput` text inputs.
        This component allows you to take inputs from users.
        
        Parameters
        ----------
        name: `str`
            The title that shows up on the modal.
            
        inputs: `list`
            A list of `TextInput` objects to be added upon creation.
            
        customID: `str`
            A developer-defined `customID` for this `Component`. If not given, this is automatically derived from the `name`.
            
        callback: `Callable`
            A callable (function, method or class) to be executed once a response has been given.
        '''
        
        if type(inputs) not in (list, tuple, set):
            
            inputs = []
        
        if name is None:
            
            name = 'Modal'
            
        if customID is None:
            
            customID = name.lower().replace(' ', '_')
            
        self.bot = None
        self.name: str = name
        self.customID: str = customID
        self.callbacks: list[Callable] = []
        self.inputs: list[TextInput] = []
        
        if callback and callable(callback):
            
            self.bind(callback)
            
        for input in inputs:
            
            if type(input) is TextInput:
                
                self.inputs.append(input)
            
    def connect(self, callback):
        
        '''
        Add an on clicked callback for this `Modal`. When someone on Discord sends something through this `Modal`, the connected `callback` will be called.
        '''
        
        if callable(callback):
            
            self.callbacks.append(callback)
            
    def bind(self, *callbacks):
        
        '''
        Similarly to `Modal.connect()`, this allows you to add another callback. The only difference is that you can pass multiple callbacks at once here.
        '''
        
        for callback in callbacks:
            
            self.connect(callback)
            
    def callback(self, func = None):
        
        '''
        You can use this as a decorator for your callback instead of using `Modal.connect` or `Modal.bind`.
        '''
        
        if func is not None and callable(func):
            
            self.bind(func)
            
    def addInput(self, input: TextInput):
        
        '''
        Add another `TextInput` input to this `Modal`.
        '''
        
        if type(input) is not TextInput:
            
            return None
        
        self.inputs.append(input)
        
    def createInput(
        self,
        name: str = None,
        style: TextInputStyle = TextInputStyle.Short,
        minLength: int = None,
        maxLength: int = None,
        placeholder: str = None,
        required: bool = True,
        customID: str = None,
        callback: Callable = None
    ):
        
        '''
        Create a new `TextInput`, add it to the `Modal`, and return it.
        '''
        
        input = TextInput(name, style, minLength, maxLength, placeholder, required, customID)
        input.connect(callback=callback)
        self.addInput(input=input)
        return input
        
        
    def input(
        self,
        name: str = None,
        style: TextInputStyle = TextInputStyle.Short,
        minLength: int = None,
        maxLength: int = None,
        placeholder: str = None,
        required: bool = True,
        customID: str = None
    ):
        
        '''
        A decorator that creates a new `TextInput` with the given arguments and the decorated function as its callback, and adds it to the `Modal`.
        '''
        
        def decorator(function: Callable) -> TextInput:
            
            nonlocal name, style, minLength, maxLength, placeholder, required, customID
            
            if not callable(function):
                
                return None
            
            return self.createInput(name, style, minLength, maxLength, placeholder, required, customID, function)
        
        return decorator
            
            
    def setBot(self, bot):
        
        '''
        For internal use only. Updates the `bot` value for this `Modal` and all of its `TextInput` input components.
        '''
        
        self.bot = bot
        
        for input in self.inputs:
            
            input.bot = bot
            
    def toJson(self):
        
        data = {
            'title': self.name,
            'custom_id': self.customID,
            'components': [inp.toJson() for inp in self.inputs]
        }
        
        return data
            
    async def __call__(self, interaction: Interaction, *args, **kwargs):
        
        '''
        Execute all bound/connected callback functions as if an option of the select menu has been clicked.
        '''
        
        from .__core__.addons import Utils
        
        coros = []
        
        for callback in self.callbacks:
            
            coros.append(
                
                Utils.execute(callback, interaction, *args, self.bot, **kwargs)
                
            )
            
        await asyncio.gather(*coros)
    
    def __str__(self):
        
        return f'<Modal name="{self.name}" customID="{self.customID}" inputs={self.inputs}>'
    
    def __repr__(self):
        
        return self.name
            

    
class View(Registrable):
    
    ACTION_ROW = 1
    BUTTON = 2
    STRING_SELECT = 3
    TEXT_INPUT = 4
    USER_SELECT = 5
    ROLE_SELECT = 6
    MENTIONABLE_SELECT = 7
    CHANNEL_SELECT = 8
    
    VALID_COMPONENT_TYPES = [1, 2, 3, 4, 5, 6, 7, 8]
    SELECT_MENU = [3, 5, 6, 7, 8]
    
    def __init__(self, name: str = ..., permanent: bool = False):
        
        '''
        Creates a new `View` with the given components in it. Views can be made up of `Button` and `SelectMenu` objects.

        Parameters
        ----------
        name: `str`
            A name that you will use to recognize your `View`. You can use this later to get it using `Bot.loadView()`.

        permanent: `bool`
            Whether the `View` should live forever. If this is set to `False`, the view will be no longer functional after not being used for 24 hours.
        '''
        
        self.rows: list[ActionRow[Component]] = [ActionRow(x + 1) for x in range(5)]
        
        for row in self.rows:
            
            row.underlying = self
        
        self.bot = None
        self.messages: list[Message] = []
        self.last: float = time.time()

        self.permanent: bool = permanent
        self.name = name

    @property
    def name(self) -> str:
        return self.__viewname__
    
    @name.setter
    def name(self, new: str):

        if type(new) is not str:
            if new not in (..., None):
                new = str(new)

        if self.__checkName__(new):
            self.__viewname__ = new

    def __checkName__(self, name: str) -> bool:

        if self.bot is None:
            return True
        
        for view in self.bot.__gateway__.views:

            if view.name == name:

                raise NameError("Can't have two or more views of the same name")
            
        return True

    def __register__(self, bot, name: str):

        from .bot import Bot
        bot: Bot = bot
        gw = bot.__gateway__
        gw.views.append(self)
        self.bot: Bot = bot
        
        if self.name in (..., None):
            self.name = name

        for child in self.children:
            if hasattr(child, "setBot"):
                child.setBot(bot)
            elif hasattr(child, "bot"):
                child.bot = bot

        self.bot.__gateway__.__updateSavedViews__()

    def __unregister__(self):

        if self.bot is None:
            return
        
        try:
            self.bot.__gateway__.views.remove(self)
        except:
            ...

        self.bot = None

            
    @property
    def components(self):
        
        '''
        All the components in this `View`, from all `ActionRow` rows, in one list.
        '''
        
        all: list[Component] = []
        
        for row in self.rows:
            
            all.extend(
                row.components
            )
            
        return all
    
    @property
    def buttons(self) -> list[Button]:
        
        '''
        All the buttons in this `View`, from all `ActionRow` rows. This is the same as `View.components`, but this contains buttons only.
        '''
        
        return [component for component in self.components if component.type == View.BUTTON]
    
    @property
    def selectMenus(self) -> list[SelectMenu]:
        
        '''
        All the select menus in this `View`, from all `ActionRow` rows.
        '''
        
        return [component for component in self.components if component.type in View.SELECT_MENU]
    
    @property
    def length(self):
        
        '''
        Returns an integer representing the amount of all components inside this `View`.
        '''
        
        return len(self.components)
    
    @property
    def children(self):

        '''
        Returns all of the View's children, whether those are `ActionRow` rows or any type of `Component`.
        '''

        return self.components + self.rows
    
    @property
    def expired(self):

        '''
        Whether more than a day has passed since the last interaction.
        '''

        return self.last > time.time() + 86400
    
    @property
    def alive(self):

        '''
        Whether this `View` is still alive. To prevent it from dying, set `View.permanent` to `True`.
        '''

        return self.expired if not self.permanent else True
        
    def addComponent(self, component: Component):
        
        '''
        Adds a new `Component` to the best available `ActionRow` and returns it. If no action row is available or the given row is full, raises an `IndexError`. Another exception that may be raised is `KeyError`, when a component with the same custom ID already exists.
        '''
        
        if not self.isUnique(component.customID):
            
            raise KeyError('Component customID must be unique to its view')
        
        if component.row is None or type(component.row) is not int:
        
            for row in self.rows:
                
                try:
                    
                    row.addComponent(component)
                    component.underlying = self
                    #component.row = row.row
                    return row
                    
                except:
                    
                    continue
                
            raise IndexError('No action row available')
        
        else:
            
            if component.row > 5 or component.row > 1:
                
                raise IndexError('Invalid action row')
            
            try:
                
                index = component.row - 1
                row = self.rows[index]
                row.addComponent(component)
                component.underlying = self
                return row
            
            except:
                
                raise IndexError('Action row does not have enough space')
    
    def removeComponent(self, component: Component):
        
        '''
        Removes an already added `Component` from its corresponding `ActionRow` and returns the action row. Raises `ValueError` if the given `Component` is not present.
        '''
        
        for row in self.rows:
            
            try:
                
                row.removeComponent(component)
                component.underlying = None
                return row
            
            except:
                
                continue
            
        raise ValueError('Component not in any action rows')
    
    def addComponents(self, *components: Component):
        
        '''
        Add all given `Component` objects to this `View`, ignoring any errors that would normally be raised by using just `addComponent`.
        '''
        
        for component in components:
            
            try:
                self.addComponent(component)
                
            except:
                continue
            
    def removeComponents(self, *components: Component):
        
        '''
        Remove all given `Component` objects from this `View`, ignoring any errors that would normally be raised by using just `removeComponent`.
        '''
        
        for component in components:
            
            try:
                self.removeComponent(component)
                
            except:
                continue
            
    def getComponent(self, customID: str, default = None):
        
        '''
        Gets a `Component` by its `customID` which should have been passed when it was created. If not found, `default` will be returned.
        '''
        
        for component in self.components:
            
            if customID == component.customID:
                
                return component
            
        return default
    
    def getButton(self, customID: str, default = None):
        
        '''
        Gets a `Button` by its `customID`. This is the same as using `getButton`, but is limited to `Button` objects and no other component types.
        '''
        
        for button in self.buttons:
            
            if customID == button.customID:
                
                return button
            
        return default
    
    def getSelectMenu(self, customID: str, default = None):
        
        '''
        Gets a `SelectMenu` by its `customID`. This is the same as using `getButton`, but is limited to `SelectMenu` objects and no other component types.
        '''
        
        for select in self.selectMenus:
            
            if customID == select.customID:
                
                return select
            
        return default
    
    def isButtonUnique(self, customID: str):
        
        '''
        Get whether the `customID` given is unique and not used for any other buttons.
        '''
        
        return self.getButton(customID) is None
    
    def isSelectMenuUnique(self, customID: str):
        
        '''
        Get whether the `customID` given is unique and not used for any other select menus.
        '''
        
        return self.getSelectMenu(customID) is None
    
    def exists(self, customID: str):
        
        '''
        Returns whether a `Component` which goes by this `customID` already exists.
        '''
        
        return self.getComponent(customID) is not None
    
    def isUnique(self, customID: str):
        
        '''
        Returns whether a component `customID` is unique and not used by any other components.
        '''
        
        return not self.exists(customID)
            
    def createComponent(self, component: type[Button | SelectMenu], *args, **kwargs):
        
        '''
        Create a new `Component` of given `component` type with the applied `*args` arguments and `**kwargs` keyword arguments, adding it to the view and returning it.
        '''
        
        if component not in (Button, SelectMenu):
            
            return None
        
        new = component(*args, **kwargs)
        self.addComponent(new)
        return new
    
    def addButton(self, name: str = None, customID: str = None, style: int = 1, emoji: str | int = None, row: int = None) -> Button:
        
        '''
        Create a new `Button`, add it to the `View`, and return it just in case it's needed for later use.
        '''
        
        return self.createComponent(
            
            component = Button,
            name = name,
            customID = customID,
            style = style,
            emoji = emoji,
            row = row
            
        )
        
    def addSelect(self, name: str = None, options: list[SelectMenuOption] = None, customID: str = None, minvalues: int = 1, maxvalues: int = 1, row: int = None) -> SelectMenu:
        
        '''
        Create a new `SelectMenu`, add it to the `View`, and return it.
        '''
        
        return self.createComponent(
            
            component = SelectMenu,
            options = options,
            name = name,
            customID = customID,
            minvalues = minvalues,
            maxvalues = maxvalues,
            row = row
        )
        
    def button(self, name: str = None, customID: str = None, style: int = 1, emoji: str | int = None, row: int = None):
        
        '''
        A decorator that creates a new `Button` with the decorated function as its `Callback`, adds it to the `View`, and returns it.
        '''
        
        def decorator(decorated: Callable) -> Button:
            
            nonlocal self, name, customID, style, row
        
            button = self.addButton(
                
                name = name,
                customID = customID,
                style = style,
                emoji = emoji,
                row = row
                
            )
            
            button.bind(decorated)
            return button
            
        return decorator
    
    def select(self, name: str = None, options: list[SelectMenuOption] = None, customID: str = None, minvalues: int = 1, maxvalues: int = 1, row: int = None):
        
        '''
        A decorator that creates a new `SelectMenu` with the decorated function as its `Callback`, adds it to the `View`, and returns it.
        '''
        
        def decorator(decorated: Callable) -> SelectMenu:
            
            nonlocal self, name, customID, row
        
            select = self.addSelect(
                
                options = options,
                name = name,
                customID = customID,
                minvalues = minvalues,
                maxvalues = maxvalues,
                row = row
                
            )
            
            select.bind(decorated)
            return select
            
        return decorator
    
    def extend(self, *views):
        
        '''
        Extends this `View` by adding as many `Component` components from the other `views` to it as possible.
        '''
        
        views = [view for view in views if type(view) is View]
        
        for view in views:
            
            for row in self:
                
                row.extend(view.rows)
            
    async def send(
        self,
        ctx: Context | Interaction = ...,
        channel: Channel = ...,
        *args, **kwargs
    ) -> Message | None:
        
        '''
        Sends the `View` as a `Message` with the given `ctx` as a `Context` or an `Interaction`. Instead of passing some context, you can pass a `channel` to send this to. If none of these parameters is passed or the types are invalid, nothing will happen.
        You can use arguments and keyword arguments to be used when sending the message, eg `content = "This is a view"`
        '''
        
        if ( ctx in (..., None) and channel in (..., None) ) or ( not ctx and not channel ):
            
            return None
        
        if channel and type(channel).__name__ != 'Channel':
            
            args += (channel,)
            channel = None
        
        if ctx and ctx not in (..., None):
            
            try:
                
                return await ctx.send(*args, **kwargs, view=self)
                
            except:
                
                return None
            
        if channel and channel not in (..., None):
            
            try:
                
                return await channel.send(*args, **kwargs, view=self)
            
            except:
                
                return None
                
    
    def toJson(self):
        
        '''
        Converts this `View` and all of its components and their sub-components into a json-like `dict` dictionary.
        '''

        if self.bot in (None, ...):

            raise NotRegisteredError(self)
        
        jsonified = [
            
            row.toJson()
            for row in self.rows
            if row.components
            
        ]
        
        return jsonified
    
    def save(self):

        """
        Saves this `View` and its children to the bot's database.
        """

        db = self.bot.__gateway__.db

        if not db.viewExists(self.name):

            db.createView(self.name)

        self.update("msgs", [msg.id if type(msg) is Message else msg for msg in self.messages])
        self.update("nm", self.name)
        self.update("lst", self.last)

    def update(self, item, value):

        self.bot.__gateway__.db.updateView(self.name, item, value)

    def updateLast(self):

        self.last = time.time()
        self.save()

    def updateMessages(self, new: Message):

        self.messages.append(new)
        self.save()

    def delete(self):

        """
        Deletes the permanent `View` from your DataBase (lunardata.db) for whatever reason.
        """

        self.bot.__gateway__.db.deleteView(self.name)

    @property
    def messageIDs(self):
        return [message.id if type(message) is Message else int(message) for message in self.messages]
    
    def __str__(self):
        
        return f'<View name="{self.name}" components={self.length}>'
    
    def __repr__(self):
        return str(self)
    
    def __iter__(self):
        
        return self.rows.__iter__()
    
    def __len__(self):

        return self.length
    
def button(name: str = None, customID: str = None, style: int = 1, emoji: str | int = None, row: int = None, view: View = None):
    
    '''
    A decorator that creates a new `Button` with the decorated function as its `Callback` and returns it.
    '''
    
    def decorator(decorated: Callable) -> Button:
        
        nonlocal name, customID, style, row
    
        return Button(
            
            name = name,
            customID = customID,
            style = style,
            emoji = emoji,
            row = row,
            callback = decorated
            
        )
        
    return decorator


from .addons import Utils, SlashOption, time
from ..basetypes import User, Interaction, Context
from ..errors import UnregisterError
 
class convertibleToStr: pass
class function: pass
class method: pass

ON_START = 0
ON_READY = 0
ON_MESSAGE = 1
ON_CHANNEL_UPDATE = 2
ON_TYPING_START = 3
ON_PRESENCE_UPDATE = 4
ON_MESSAGE_DELETE = 5
ON_MESSAGE_EDIT = 6
ON_MESSAGE_UPDATE = 6
ON_REACTION_ADD = 7
ON_REACTION_REMOVE = 8
ON_TYPING_STOP = 9
ON_CHANNEL_CREATE = 10
ON_CHANNEL_DELETE = 11
ON_MEMBER_JOIN = 12
ON_MEMBER_UPDATE = 13
ON_MEMBER_LEAVE = 14


class Bot: ... # Can't import lunarcord.Bot so why not..

class Registrable:
    def __init__(self):
        
        self.bot: Bot = None
        self.__registered__: bool = False
        
    def __register__(self, bot: Bot):
        '''
        Implement `bot.register(self)`.
        
        This should generally be something like:
        ```
        def __register__(self, bot: lunarcord.Bot):
            
            self.bot = bot
            name = "whateverName"
            
            if not hasattr(self.bot, name) or type(getattr(self.bot, name)) is not list:
                setattr(self.bot, name, [])
                
            items = getattr(self.bot, name)
            items.append(self)
            setattr(self.bot, name, items)
            self.__registered__ = True
            
        ```
        The above simply gets a list named `name` (in this case "whateverName") for the bot (or creates it if it doesn't exist) and adds `self` `(lunarcord.Registrable)` to the list.
        '''
        
        self.bot = bot
        name = "registrations"
        if not hasattr(self.bot, name) or type(getattr(self.bot, name)) is not list:
            setattr(self.bot, name, [])
        items: list = getattr(self.bot, name)
        items.append(self)
        setattr(self.bot, name, items)
        
    def __unregister__(self):
        '''
        Implement `bot.unregister(self)`.
        
        This should generally be something like:
        ```
        def __unregister__(self):
            
            name = "whateverName"
            
            if not hasattr(self.bot, name) or type(getattr(self.bot, name)) is not list:
                setattr(self.bot, name, [])
            items = getattr(self.bot, name)
            
            if self not in items:
                print("Failed to unregister item - it hasn't been registered before!")
            else:
                items.remove(self)
                setattr(self.bot, name, items)
                self.__registered__ = False
                
            self.bot = None
            
        ```
        The above simply gets a list named `name` (in this case "whateverName") for the bot (or creates it if it doesn't exist) and removes `self` `(lunarcord.Registrable)` to the list - or prints an error if it hasn't been registered before..
        '''
        
        name = "registrations"
        
        if not hasattr(self.bot, name) or type(getattr(self.bot, name)) is not list:
            setattr(self.bot, name, [])
            
        items: list = getattr(self.bot, name)
        if self not in items:
            raise UnregisterError()
        
        items.remove(self)
        setattr(self.bot, name, items)
        
        self.bot = None
        
    def __reregister__(self):
        '''
        Implement `bot.reregister(self)`.
        
        This should generally be something like:
        ```
        def __reregister__(self):
            if self.__registered__:
                bot = self.bot
                self.__unregister__()
            self.__register__(bot)
        ```
        '''
        
        bot = self.bot
        self.__unregister__()
        self.__register__(bot)
        
    @property
    def registered(self):
        '''
        Whether the registrable is currently registered to a bot.
        '''
        
        return self.bot is not None
    
class TextCommand(Registrable):
    
    def __init__(self, bot: Bot, name: str | convertibleToStr, description: str | convertibleToStr, callback: function | method, aliases: list[str], cooldown: int | float):
        
        self.bot = bot
        self.name = str(name).strip()
        self.names = [self.name] + aliases
        self.namesLower = [n.lower() for n in self.names]
        self.description = str(description).strip().replace('\n', ' ')
        self.callback = callback

        if cooldown not in (None, ...):
            self.cooldown = float(cooldown)
            self.hasCooldown = True

        else:
            self.cooldown = None
            self.hasCooldown = False
        
        self.__addedToBot__ = False
        
        if self.bot is not None:
            self.bot.textCommands.append(self)
            self.__addedToBot__ = True
            
    def __eq__(self, other):
        
        if not isinstance(other, TextCommand):
            return False
        
        if self.name != other.name:
            return False
        
        if self.description != other.description:
            return False
        
        return True
            
    def __register__(self, bot: Bot, name: str = ...):
        
        self.bot = bot
        self.bot.textCommands.append(self)
        self.__addedToBot__ = True
        
    def __unregister__(self):
        
        self.bot.textCommands.remove(self)
        self.bot = None
        self.__addedToBot__ = False
        
    def __reregister__(self):
        
        bot = self.bot
        
        if bot is not None:
            self.__unregister__()
            self.__register__(bot)
            
    def __repr__(self):
        
        return f'<TextCommand name="{self.name}" description="{self.description}"'
    
    def __str__(self):
        
        return self.name
    
    async def __call__(self, ctx: Context, *args, **kwargs):
        '''
        Call the text command's callback function, applying the context in which it was triggered as well as the args and kwargs given.
        '''
        
        if len(args) == 1:
            args = [args[0]]
            
        else:
            args = list(args)

        await Utils.execute(
            self.callback,
            ctx,
            *args,
            **kwargs,
            bot = self.bot
        )

        if self.hasCooldown:

            ends = self.cooldown + time.time()

            ctx.author.setCooldown(
                name=self.name,
                ends=ends,
                type="textcommand"
            )

class SlashCommand(Registrable):
    
    def __init__(self, bot, name: str | convertibleToStr, description: str | convertibleToStr, guilds: list[int], callback: function | method, cooldown: int | float):
        
        self.bot = bot
        self.name = str(name).strip()
        self.description = str(description).strip().replace('\n', ' ')
        self.callback = callback
        self.options = list(self.getOptions())
        self.id: int = None
        self.cooldown: int | float = cooldown
        
        if guilds is None:
            self.guilds = guilds
            
        else:
            self.guilds: list[int] = [int(guild) for guild in guilds]
        
        self.__addedToPending__ = False
        self.__addedToGateway__ = False
        
        if self.bot is not None:
            self.bot.pendingSlashCommands.append(self)
            self.__addedToPending__ = True
            
    def __register__(self, bot: Bot, name: str = ...):
        
        self.bot = bot
        self.bot.pendingSlashCommands.append(self)
        self.__addedToPending__ = True
        
    def __unregister__(self):
        
        try:
            self.bot.pendingSlashCommands.remove(self)
        except:
            pass
        self.bot = None
        self.__addedToPending__ = False
        
    def __reregister__(self):
        
        bot = self.bot
        
        if bot is not None:
            self.__unregister__()
            self.__register__(bot)
        
    def getOptions(self):
        
        if True:
            from ..basetypes import User, Channel, Role, Member
            from ..builders import Attachment
            Types = (User, Channel, Role, Member, "Mentionable", Attachment)
                
        options, self.params = Utils.slashOptions(
            function = self.callback,
            Types = Types
        )
        
        return options
    
    async def toJson(self, override: bool = True):
        
        data: dict = await self.bot.manager.createSlash(
            
            name = self.name,
            description = self.description,
            options = self.options,
            override = override,
            guilds = self.guilds
            
        )
        
        if override:
        
            if self.id is None:
                
                self.id = data.get('id')
            
            if self.bot is not None:
                
                self.bot.__gateway__.slashCommands.append(self)
                self.__addedToGateway__ = True
        
        return data
        
    def fromJson(bot, data, cooldown: int | float = None):
        
        new = SlashCommand(
            bot = bot,
            name = data['name'],
            description = data['description'],
            options = data['options'],
            guilds = [data.get('guild_id')],
            cooldown = cooldown
        )
        
        return new
        
    async def __call__(self, interaction: Interaction, *args, **kwargs):
        '''
        Call the slash command's on-interaction callback function, applying the interaction that triggered it as well as the arguments given.
        '''
        
        return await Utils.execute(
            self.callback,
            interaction,
            *args,
            **kwargs,
            bot = self.bot,
            params = self.params,
        )
    
    def __repr__(self):
        return f'<SlashCommand name="{self.name}" description="{self.description}">'
    
    def __str__(self):
        return self.name
    
class Event(Registrable):
    def __init__(self, bot, name: str | convertibleToStr, type: int, callback: function | method, container: str, params: dict):
        
        self.bot = bot
        self.name = name
        self.type = type
        self.callback = callback
        self.container = container
        self.extraparams = params
        self.__connected__ = False
        
        if self.bot is not None and not self.__connected__:
            self.connect()
            
    def __register__(self, bot: Bot, name: str = ...):
        
        self.bot = bot
        self.connect()
        self.__connected__ = True
        
    def __unregister__(self):
        
        self.deconnect()
        self.bot = None
        self.__connected__ = False
        
    def __reregister__(self):
        
        bot = self.bot
        
        if bot is not None:
            self.__unregister__()
            self.__register__(bot)
        
    def connect(self):
        new_container = getattr(self.bot.__gateway__, self.container)
        new_container.append(self)
        setattr(self.bot.__gateway__, self.container, new_container)
        self.connected = True
        
    def deconnect(self):
        new_container = getattr(self.bot.__gateway__, self.container)
        try:
            new_container.remove(self)
        except:
            ...
        setattr(self.bot.__gateway__, self.container, new_container)
        self.connected = False
        
    async def __call__(self, *args, **kwargs):
        '''
        Emit the event and execute its callback, applying the given arguments.
        '''
        
        await Utils.execute(
            self.callback,
            *args,
            **kwargs
        )
        
    def __repr__(self):
        return f'<Event name="{self.name}" type="{self.type}" callback="{self.callback}">'
    
    def __str__(self):
        return self.name


import asyncio, aiohttp
from urllib.parse import quote
from ..errors import RequestError, UnauthorizedError

class Manager:
    
    aiohttp = aiohttp
    requests = ...
    
    def __init__(self, token: str = None, session: aiohttp.ClientSession = None, isBot: bool = True, gateway = None):
        
        if session is None:
            
            session = aiohttp.ClientSession()
        
        self.token = token
        self.isBot = True
        self.headers = {'Authorization': None, 'Content-Type': 'application/json'}
        self.session = session
        self.gateway = gateway
        self.generateHeaders()

        self.reason: str = ...
        
    async def close(self):
        
        await self.session.close()
        
        
    def generateHeaders(self, token: str = None):
        
        if token is not None:
            self.token = token
            
        self.headers = {
            'Authorization': self.token,
            'Content-Type': 'application/json'
		}
        
        return self.headers
    
    async def request(self, url: str, json: dict = None, files: list = None, type: str = 'get', version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False):
        
        if self.session.closed:
            
            return None
        
        headers = self.headers
        reason = self.reason
        
        if "Authorization" not in headers:
            if self.token is not None:
                self.generateHeaders(self.token)
                
            else:
                disableHeaders = True

        if reason not in (..., None):
            self.headers["X-Audit-Log-Reason"] = str(reason)
            self.reason = None
        
        if disableHeaders:
            headers = {}
            
        type = type.strip().lower()
        
        if type not in ('get', 'post', 'delete', 'put', 'patch'):
            raise RequestError(None, 'Invalid method used', type)
        
        if not customUrl:
            url = f'https://discord.com/api/v{version}/' + url
        
        kwargs = {
            
            'headers': headers,
            'json': json,
            'params': params
            
        }
        
        async with self.session.request(type, url, **kwargs) as response:

            self.headers.pop("X-Audit-Log-Reason", None)
            code = int(response.status)
            
            successful = True
            
            if code in (400, 401, 403, 404, 405, 429, 502, '5xx'):
                successful = False

            errdict = {}
                
            if not successful:
                
                try:
                    
                    try:
                        errdict = await response.json()
                        msg = errdict['message']
                    except:
                        msg = None
                    
                except:

                    errdict = {}
                    
                    match code:
                        
                        case 400:
                            msg = 'Bad Request'
                            
                        case 401:
                            msg = 'Unauthorized'
                            
                        case 403:
                            msg = 'Forbidden'
                            
                        case 404:
                            msg = 'Not found'
                            
                        case 405:
                            msg = 'Method not allowed'
                            
                        case 429:
                            msg = 'You are being rate limited'
                            
                        case 502:
                            msg = 'Gateway unavailable'
                            
                        case '5xx':
                            msg = 'Discord Error'
                            
                        case _:
                            msg = 'Unknown Error'
                            
                if code == 429:
                    
                    slowdown = int(errdict.get('retry_after', 10))
                    #print("Slowdown! Time:", slowdown)
                    await asyncio.sleep(slowdown)
                    return await self.request(url, json, files, type, version, returns, customUrl, params, disableHeaders)
                
                elif code == 401:
                    
                    raise UnauthorizedError()
                    
                else:
                    
                    if msg is not None:
                        raise RequestError(code, msg, type)
                    
                    else:
                        raise RequestError(code, f"Unknown error: {code}")
                
                
            
            if isinstance(returns, str):
                returns = returns.lower()
            
            if returns in ('raw', 'bytes', 'default', bytes):
                return await response.read()
            
            elif returns in ('json', 'dict', 'dictionary', dict):
                return await response.json()
            
            elif returns in ('content', 'str', 'string', str):
                return await response.text()
            
            elif returns in ('code', 'status', 'statuscode', int):
                return response.status
            
            elif returns in ('none', 'null', 'empty', None):
                return None
            
            elif returns in ('result', 'success', 'succesfull', bool):
                return True
            
            else:
                raise RequestError(0, 'Bad request')
        
    async def get(self, url: str, json: dict = None, files: list = None, version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False): return await self.request(url, json, files, 'get', version, returns, customUrl, params, disableHeaders)
    async def post(self, url: str, json: dict = None, files: list = None, version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False): return await self.request(url, json, files, 'post', version, returns, customUrl, params, disableHeaders)
    async def delete(self, url: str, json: dict = None, files: list = None, version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False): return await self.request(url, json, files, 'delete', version, returns, customUrl, params, disableHeaders)
    async def put(self, url: str, json: dict = None, files: list = None, version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False): return await self.request(url, json, files, 'put', version, returns, customUrl, params, disableHeaders)
    async def patch(self, url: str, json: dict = None, files: list = None, version: int = 10, returns: type = dict, customUrl: bool = False, params: dict = {}, disableHeaders: bool = False): return await self.request(url, json, files, 'patch', version, returns, customUrl, params, disableHeaders)

    async def loadMessages(self, channel: int, limit: int = 1000) -> list[dict]:

        params = {
            "limit": limit
        }
        
        url = f'channels/{channel}/messages'
        return await self.get(url, params=params)
        
    async def loadMessage(self, channel: int, message: int) -> dict | None:
        
        url = f'channels/{channel}/messages/{message}'
        
        try:
            return await self.get(url)
            
        except:
            return await self.loadMessage2(channel, message)
    
    async def loadMessage2(self, channel: int, message: int) -> dict | None:
        
        messages = await self.loadMessages(channel)
        
        for log in messages:
            
            id = int(log.get('id'))
            
            if int(message) == id:
                return log
            
			
    async def sendMessage(self, channel: int, content: str, reference: int = None, embeds: list = [], attachments: list = [], components: list = []) -> dict:

        if not isinstance(content, str) and content is not None:
            content = str(content)
        
        if ( content is None or content.strip() == '' ) and ( not embeds and not attachments and not components ):
            content = '** **'
        
        url = f'channels/{channel}/messages'
        
        data = {'content': content, 'allowed_mentions': None}
        
        if reference is not None:
            data['message_reference'] = {'message_id': reference}
        
        if embeds:
            
            jsonEmbeds: list[dict] = []
            
            for embed in embeds:
                
                jsonEmbeds.append(await embed.__json__())
                
            data['embeds'] = jsonEmbeds
            
        if components:
            
            data['components'] = components
            
        return await self.post(url, data, attachments)
    
    async def loadUser(self, id: int):
        url = f'users/{id}'
        return await self.get(url)
        
    async def addReaction(self, channel: int, message: int, reaction: str) -> None:
        
        try:
            reaction = str(reaction)
        except:
            return None
        
        reaction = quote(reaction, safe="")
        url = f'channels/{channel}/messages/{message}/reactions/{reaction}/@me'
        return await self.put(url)
        
    async def removeReaction(self, channel: int, message: int, reaction: str) -> None:
        
        try:
            reaction = str(reaction)
        except:
            return None
        
        reaction = quote(reaction, safe="")
        url = f'channels/{channel}/messages/{message}/reactions/{reaction}/@me'
        return await self.delete(url)
        
    async def deleteMessage(self, channel: int, message: int) -> None:
        
        url = f'channels/{channel}/messages/{message}'
        return await self.delete(url)
        
    async def me(self) -> dict | None:
        
        return await self.get('users/@me')
    
    async def clientInfo(self) -> dict | None:
        
        return await self.get('oauth2/applications/@me')
    
    async def clientID(self) -> int:
        
        info = await self.clientInfo()
        return int(info.get('id'))
    
    async def editMessage(self, channel: int, message: int, content: str) -> None:
        url = f'channels/{channel}/messages/{message}'
        data = {'content': str(content)}
        return await self.patch(url, data)
    
    async def loadRoles(self, guild: int) -> dict:
        url = f"guilds/{guild}/roles"
        return await self.get(url)
    
    async def addRole(self, guild: int, user: int, role: int, reason: str = ...):
        self.reason = reason
        url = f"/guilds/{guild}/members/{user}/roles/{role}"
        await self.put(url, returns=None)
    
    async def removeRole(self, guild: int, user: int, role: int, reason: str = ...):
        url = f"/guilds/{guild}/members/{user}/roles/{role}"
        self.reason = reason
        await self.delete(url, returns=None)
    
    async def kickMember(self, guild: int, user: int):
        url = f"guilds/{guild}/members/{user}"
        await self.delete(url, returns=None)

    async def banMember(self, guild: int, user: int, deleteMessageSeconds: int = 0):
        data = {"delete_message_seconds": deleteMessageSeconds}
        url = f"guilds/{guild}/bans/{user}"
        await self.put(url, json=data, returns=None)
    
    async def createSlash(self, name: str = None, description: str = None, options: list = [], type: int = 1, guilds: list[int] = None, override: bool = True) -> dict:
        
        data = {
		   'name': name,
		   'description': description,
		   'type': 1, # Slash Command
		   'options': options
		}
        
        if not override:
            return data
        
        clientID = await self.clientID()
        
        if guilds is not None and len(guilds) > 0:
            
            try:
                guildID = guilds.pop(0)
                url = f'applications/{clientID}/guilds/{guildID}/commands'
                last = await self.post(url, data)
                last = await self.createSlash(name, description, options, type, guilds, override)
                
            except:
                return last
            
        else:
        
            url = f'applications/{clientID}/commands'
            return await self.post(url, data)
        
        return last
    
    async def slashCommands(self) -> list[dict] | None:
        
        clientID = await self.clientID()
        
        slash: list = await self.get(
            f'applications/{clientID}/commands'
        )
        
        return slash
    
    async def deleteSlash(self, id: int | str) -> None:
        
        clientID = await self.clientID()
        
        await self.delete(
            f'applications/{clientID}/commands/{id}',
            returns = None
        )
        
    async def deleteSlashCommands(self, *ids: int | str) -> None:
        '''
        Deletes the commands with the specified IDs, or all if no IDs are given.
        
        Parameters
        ----------
        ids: `int`, `str`
            A list of IDs of the Slash Commands to delete. 
        '''
        
        if len(ids) <= 0:
            slashCommands = await self.slashCommands()
            
        else:
            slashCommands = [{'id': slashid} for slashid in ids]
            
        async def delete(slashCommand):
        
            try:
                
                await self.deleteSlash(
                    id = slashCommand['id']
                )
                
            except:
                
                ...
            
        coros = []
        
        for slashCommand in slashCommands:
            
            coros.append(delete(slashCommand))
            
        await asyncio.gather(*coros)
	   
    async def interaction(self, id) -> dict:
        
        url = f'interactions/{id}'
        return await self.get(url)
    
    async def respondToInteraction(self, interaction, payload, attachments: dict = None) -> None:
        
        url = f'interactions/{interaction.interactionID}/{interaction.interactionToken}/callback'
        
        return await self.post(
            url = url,
            json = payload,
            files = attachments,
            returns = None
        )
		
    async def getSettings(self) -> dict:
        
        url = 'users/@me/settings'
        return await self.get(url)
		
    async def setSettings(self, data) -> None:
        
        url = 'users/@me/settings'
        return await self.patch(url, data)
		
    async def loadChannel(self, id: int) -> dict:
        
        url = f'channels/{id}'
        return await self.get(url)
    
    async def loadChannelsFor(self, guild: int) -> list[dict]:
        
        return await self.get(
            url = f'guilds/{guild}/channels'
        )
        
    async def loadChannels(self) -> list[dict]:
        
        return await self.get(
            url = 'users/@me/channels'
        )

    async def updateChannel(self, id: int, data) -> None:
        
        url = f'channels/{id}'
        return await self.patch(
            url=url, json=data
        )

    async def setChannelName(self, id: int, name: str) -> None:
        
        data = {
            'name': name
        }
        
        await self.updateChannel(id, data)
		
    async def pinMessage(self, channel: int, id: int) -> None:
        
        url = f'channels/{channel}/pins/{id}'
        return await self.put(url)
    
    async def unpinMessage(self, channel: int, id: int) -> None:
        
        url = f'channels/{channel}/pins/{id}'
        return await self.delete(url)
    
    async def channelFor(self, user: int) -> dict:
        
        url = f'users/@me/channels'
        user = str(user)
        
        payload = {
            'recipient_id': user
        }
        
        return await self.post(url, payload)
    
    async def showTyping(self, channel: int) -> None:
        
        url = f'channels/{channel}/typing'
        return await self.post(url)
    
    async def dmRecipients(self, group: int) -> list[dict]:
        
        return await self.get(
            url = f'channels/{group}/recipients'
        )
    
    async def groupDmAddRecipient(self, group: int, user: int) -> None:
        
        return await self.put(
            url = f'channels/{group}/recipients/{user}'
        )
        
    async def groupDmRemoveRecipient(self, group: int, user: int) -> None:
        
        return await self.delete(
            url = f'channels/{group}/recipients/{user}'
        )
        
    async def createThreadFromMessage(self, channel: int, message: int, name: str, autoArchive: int = None, rateLimit: int = None) -> dict:
        
        if rateLimit < 0:
            rateLimit = 0
            
        if rateLimit > 21600:
            rateLimit = 21600
            
        if autoArchive not in (60, 1440, 4320, 10080):
            autoArchive = None
            
        payload = {
            'name': name,
            'auto_archive_duration': autoArchive,
            'rate_limit_per_user': rateLimit
        }
        
        return await self.post(
            url = f'channels/{channel}/messages/{message}/threads',
            json = payload
        )
        
    async def createThread(self, channel: int, name: str, autoArchive: int = None, rateLimit: int = None) -> dict:
        
        if rateLimit < 0:
            rateLimit = 0
            
        if rateLimit > 21600:
            rateLimit = 21600
            
        if autoArchive not in (60, 1440, 4320, 10080):
            autoArchive = None
            
        payload = {
            'name': name,
            'auto_archive_duration': autoArchive,
            'rate_limit_per_user': rateLimit
        }
        
        return await self.post(
            url = f'channels/{channel}/threads',
            json = payload
        )
        
    async def loadGuilds(self) -> list[dict]:
        
        return await self.get(
            url = 'users/@me/guilds'
        )
        
    async def loadGuild(self, id: int) -> dict:
        
        return await self.get(
            url = f'guilds/{id}'
        )
        
    async def guildChannels(self, id: int) -> list[dict]:
        
        return await self.get(
            url = f'guilds/{id}/channels'
        )
        
    async def login(self, emailOrPass: str, password: str) -> str:
        
        payload = {
            
            'login': emailOrPass,
            'password': password
            
        }
        
        returns: dict = await self.post(
            
            url = 'auth/login',
            json = payload,
            disableHeaders = True
            
        )
        
        return returns.get('token')
    
    async def loadMember(self, guild: int, id: int) -> dict:

        return await self.get(
            f"guilds/{guild}/members/{id}"
        )

    async def loadMembersPartial(self, guild: int, count: int = 1000, after: int = 0) -> list[dict]:

        params = {
            "limit": count,
            "after": after
        }

        return await self.get(
            f"guilds/{guild}/members",
            params = params,
            returns = "json"
        )
    
    async def loadMembers(self, guild: int):

        finished = False
        members: list[dict] = []
        highest: int = 0

        while not finished:

            try:
                
                page = await self.loadMembersPartial(guild, after=highest)

                if len(page) > 0:
                    highest = page[-1].get("user", {}).get("id", 0)
                    members.extend(page)

                else:
                    finished = True

            except:

                finished = True

        return members


PLAYING = 0
STREAMING = 1
LISTENING = 2
WATCHING = 3

ONLINE = 'online'
IDLE = 'idle'
DND = 'dnd'
DO_NOT_DISTURB = DND
DONOTDISTURB = DND
IDLING = IDLE
ACTIVE = ONLINE
INVISIBLE = 'invisible'
INVIS = INVISIBLE
OFFLINE = INVIS
INACTIVE = INVIS

from .errors import StatusError

class Activity:
    def __init__(self, type: int, name: str):
        self.type = type
        self.name = name
        
    def toJson(self):
        return {'type': self.type, 'name': self.name}
		
	
class Playing(Activity):
    def __init__(self, game: str):
        super().__init__(PLAYING, game)
		
class Watching(Activity):
    def __init__(self, video: str):
        super().__init__(WATCHING, video)
		
class Streaming(Activity):
    def __init__(self, game: str):
        super().__init__(STREAMING, game)

class Listening(Activity):
    def __init__(self, song: str):
        super().__init__(LISTENING, song)

Game = Playing
Video = Watching
Stream = Streaming
Music = Listening		
ListeningTo = Listening


'''
Lunarcord - Discord Bot Library
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A Library for easily writing multi-purpose Discord Bots.

:copyright: Copyright (c) 2024 RevoLNo.
:license: Read LICENSE for additional details.
'''

from .activities import *
from .bot import *
from .basetypes import *
from .future import *
from .builders import *

from .ui import View, Button, ButtonStyle, SelectMenu

from .__core__.types import (
    
    ON_START,
    ON_MESSAGE,
    ON_CHANNEL_UPDATE,
    ON_TYPING_START,
    ON_PRESENCE_UPDATE,
    ON_MESSAGE_DELETE,
    ON_MESSAGE_EDIT,
    ON_MESSAGE_UPDATE,
    ON_REACTION_ADD,
    ON_REACTION_REMOVE,
    
    SlashOption
)


from .errors import ParameterError
from .basetypes import Message
import asyncio

class FutureMessage:
    def __init__(self, bot, content: str, channel: int | str):
        '''
        A planned message that can be later sent to the specified channel at any moment and as many times as desired, using its :meth:`send` method.
        
        Parameters
        ----------
        bot: `Lunarcord.Bot`
            The bot client to send the message from.
        
        content: `str`
            The content of the message to send. Must be a string or convertible to string.
            
        channel: `int`, `Lunarcord.Channel`
            The channel to send the message to. Can be a discord snowflake, or a Channel object.
        '''
        
        self.content = content
        self.channel = channel
        self.bot = bot
        
    @property
    def content(self):
        '''
        The future message's content. Must be a string or convertible to string.
        '''
        
        return self.__content__
        
    @content.setter
    def content(self, value):
        if type(value) is not str:
            try: value = str(value)
            except: raise ParameterError(type(value).__name__)
            
        self.__content__ = value
        
    @content.deleter
    def content(self):
        self.__content__ = None

    @property
    def channel(self):
        '''
        The channel in which the planned message will be sent. Must be an integer or convertible to integer, representing a discord snowflake. It could also be a Channel object.
        '''
        
        return self.__channel__
    
    @channel.setter
    def channel(self, value):
        if type(value) not in (int, str):
            try: value = int(value)
            except: raise ParameterError(type(value).__name__)
            
        self.__channel__ = value
        
    async def send(self, delay: int | float = None):
        '''
        Sends the planned/future message and returns the result.
        
        Parameters
        ----------
        delay: `float`, `int`
            How much time to wait before sending the message. Defaults to None, which sends the message instantly.
            
        Returns
        -------
        message: `Lunarcord.Message`
            A Message object representing the newly sent message.
        '''
        
        if delay:
            await asyncio.sleep(delay)
            
        new = await self.bot.__gateway__.manager.sendMessage(self.channel, self.content)
        msg = Message(
            bot=self.bot,
            data=new
        )
        await msg.__proc__()
        
        return msg


from .__core__.gateway import Gateway, _, Utils, Manager, User, Channel, Message, Guild, Snowflake, Signal, Role, Member, inspect, os
from . import Activity, Playing, Watching, Listening, Streaming, StatusError
from .errors import UnauthorizedError, RegisterError, UnregisterError, ReregisterError
from .commands import __createSlashCommand__, __createEvent__, __createTextCommand__, SlashCommand, TextCommand, Registrable, Event
from .tasks import __createTask__, __createLoop__, NameLike, TimeLike, Iterable, Task, Loop

import asyncio, time, sys, json

from importlib.util import spec_from_file_location, module_from_spec
from os.path import isfile, split, join
from os import listdir

from threading import Thread

class NoTimeout: ...

class Anyone: ...
class Anything: ...
class Any: ...


class Observer(Thread):
    
    '''
    Looks and notifies for changes in given files.
    '''
    
    def __init__(self, files: list[str] = []):
        
        super().__init__()
        
        self.running: bool = True
        
        self.files: list[str] = files
        self.times: dict[str, float] = {}
        
        self.updatedPaths: list[str] = files
        self.deletedPaths: list[str] = files
        self.areDeleted: list[str] = []
        
        self.updated: Signal = Signal(str, bool)
        self.deleted: Signal = Signal(str, bool)
        
    def addFile(self, *file: str, updated: bool = True, deleted: bool = True):
        
        files = list(file)
        self.files.extend(files)
        
        if updated:
            self.updatedPaths.extend(files)
            
        if deleted:
            self.deletedPaths.extend(files)
        
        for file in files:
            
            if file in self.areDeleted: # To avoid confusion
                self.areDeleted.remove(file)
            
            if updated:
                self.setLastUpdate(file)
        
    def removeFile(self, *file: str):
        
        files = list(file)
        
        for file in files:
            
            if file in self.files:
                self.files.remove(file)
                
            if file in self.areDeleted:
                self.areDeleted.remove(file)
                
            if file in self.updatedPaths:
                self.updatedPaths.remove(file)
                
            if file in self.deletedPaths:
                self.deletedPaths.remove(file)
                
    def notify(self, *args):
        
        self.updated.callSync(*args)
        
    def getLastUpdate(self, path: str):
        
        try:
            
            timestamp = os.path.getmtime(path)
        
        except:
            
            timestamp = time.time()
            
        return timestamp
    
    def setLastUpdate(self, path: str):
            
        self.times[path] = self.getLastUpdate(path)
        
        
    def isUpdated(self, path: str):
        
        if self.times.get(path) < self.getLastUpdate(path):
            
            self.setLastUpdate(path)
            return True
        
        return False
    
    def isDeleted(self, path: str):
        
        if not os.path.exists(path):
            
            self.removeFile(path)
            return True
        
        return False
    
    def search(self, paths: list[str]):
        
        for path in paths:
            
            if path in self.updatedPaths and path not in self.areDeleted:
                
                if self.isUpdated(path):
                    self.notify(path, os.path.isfile(path))
                
    def lookForDeleted(self, paths: list[str]):
        
        for path in paths:
            
            if path in self.deletedPaths and path not in self.areDeleted:
            
                if self.isDeleted(path):
                    self.deleted.callSync(path)
                
    def run(self):
        
        while self.running:
            
            if self.files:
                self.lookForDeleted(paths=self.files)
                self.search(paths=self.files)
                
            time.sleep(1.0)
            
    def stop(self):
        
        self.running = False

class Bot:
    
    def __init__(self, prefix: str = None, prefixes: tuple | list | set = [], token: str = None):
        '''
        A Bot object representing a Discord Bot User.
        
        Parameters
        ----------
        prefix: `str`
            A prefix for all text commands. It is appended to the `prefixes` list, if provided.
            
        prefixes: `tuple`, `list`, `set`
            A tuple, list or set of multiple unique prefixes. If `prefix` is given, it will be added to the list.
        
        token: `str`
            A Discord user or bot token to connect to the gateway with. If provided, it won't be needed in `Bot.run()`.
        '''
        
        try:
            fback = inspect.currentframe().f_back
            module = inspect.getmodule(fback)
            mpath = inspect.getsourcefile(module)
            origin = os.path.dirname(mpath)
            
        except:
            origin = None
            mpath = None
        
        self.origin: str = origin # Original path
        self.mainfile: str = mpath # Main file path
        
        __this__ = sys.modules[__name__]
        path = inspect.getsourcefile(__this__)
        
        self.installation: str = os.path.dirname(path)
        
        # self.installation is the path of the current
        # lunarcord installation being used.
        
        #print(self.mainfile)
        #self.__loadMainFile()
        
        if self.origin != self.installation:
        
            #registrables = self.__loadMainFile()
            #print(registrables)
            
            ...
        
        self.startTimes = 0
        
        prefixes: list = list(prefixes)
        prefixes = [str(prefix) for prefix in prefixes]
        
        if prefix is not None:
            
            if type(prefix) is not str:
                prefix = str(prefix)
            
            prefixes.append(prefix)
            
        prefixes = Utils.removeDuplicates(prefixes)
        
        self.asyncio = asyncio
        self.__gateway__: Gateway = Gateway(self)
        self.prefixes: list = prefixes
        self.presence: dict = {}
        self.token: str = token
        
        self.pendingSlashCommands: list[SlashCommand] = []
        self.textCommands: list[TextCommand] = []
        
        #self.loadedFiles: list[str] = []
        self.registered: dict[list[Registrable]] = {}
        
        self.observer = Observer()
        self.observer.start()
        self.observer.updated.connect(self.__fileUpdated__)
        self.observer.deleted.connect(self.__fileDeleted__)
        
    
    @property
    def isBot(self):
        '''
        Whether this is a bot or normal user account.
        '''
        
        return self.__gateway__.isBot
    
    @property
    def uptime(self):
        
        '''
        The bot's uptime. (Seconds elapsed since start)
        '''
        
        return self.__gateway__.uptime
        
    @property
    def mention(self):
        
        '''
        Returns a string that can be used to mention this `Bot`, for example `<@12345678>` where `12345678` is the bot's ID.
        '''
        
        return '<@' + str(self.id) + '>'
    
    @property
    def started(self):
        
        '''
        A boolean value representing whether the `Bot` is ready. This is `False` before the `onReady` event is executed, and `True` right after.
        '''
        
        return self.__gateway__.isReady
        
    async def start(self, token: str = None, login: str = None, password: str = None, delay: float = None, terminate: float = None, loop = None):
        '''
        Initializes the bot, connects to the gateway, and starts the running process.
        
        Parameters
        ----------
        token: `str`
            The bot's token, used to connect. If provided when the `lunarcord.Bot` object was created, this parameter is not needed.
        
        login: `str`
            If `token` was never given, this will be used along with `password`. This method can only be used for normal user accounts, and this parameter should be a `str` representing an account's `email` or `phone number`.
            
        password: `str`
            Cooperates with `login` in order to get the user's token for you. This is a `str` representing a discord account's `password`. Read `login` for more details on how those two work together.
            
        delay: `float`
            Wait `delay` seconds before starting the bot. By default, do not wait any additional time.
            
        terminate: `float`
            Run for `terminate` seconds, then exit the running process automatically.
            
        loop: `asyncio.AbstractEventLoop`
            The event loop to use. If not given, `asyncio.get_event_loop()` is used instead.
        '''
        
        self.startTimes += 1
        
        if delay is not None:
            await asyncio.sleep(delay)
        
        if self.startTimes > 1: # Don't create new Gateway if it's the first time starting the bot
            self.__gateway__ = Gateway(self, self.__gateway__)
            
        if token is None and login is not None and password is not None:
            
            token = await self.login(login, password)
        
        if self.token is not None:
            
            token = self.token
            
        else:
            
            if token is None:
                raise UnauthorizedError()
            
            self.token = token
        
        self.__gateway__.setToken(token)
        
        existing = await self.manager.slashCommands()
        await self.__updateSlashCommands__(existing=existing)
                
        def terminateAfter(duration: int | float, gateway: Gateway):
            
            if duration is None:
                return None
            
            time.sleep(duration)
            gateway.stop()
            
        terminateThread = Thread(target=terminateAfter, args=(terminate, self.__gateway__))
        await self.__gateway__.start(runAfterThread=terminateThread)
        
    def run(self, token: str = None, login: str = None, password: str = None, delay: float = None, terminate: float = None):
        '''
        Initializes the bot, connects to the gateway, and starts the running process.
        
        Parameters
        ----------
        token: `str`
            The bot's token, used to connect. If provided when the `lunarcord.Bot` object was created, this parameter is not needed.
        
        login: `str`
            If `token` was never given, this will be used along with `password`. This method can only be used for normal user accounts, and this parameter should be a `str` representing an account's `email` or `phone number`.
            
        password: `str`
            Cooperates with `login` in order to get the user's token for you. This is a `str` representing a discord account's `password`. Read `login` for more details on how those two work together.
            
        delay: `float`
            Wait `delay` seconds before starting the bot. By default, do not wait any additional time.
            
        terminate: `float`
            Run for `terminate` seconds, then exit the process automatically.
        '''
        
        loop = None
        
        try:
            loop = asyncio.get_event_loop()
            
        except:
            loop = None
            
        if not loop:
            
            loop = asyncio.new_event_loop()
        
        loop.create_task(
            
            self.start(
                token = token,
                login = login,
                password = password,
                delay = delay,
                terminate = terminate
            )
            
        )
        
        loop.run_forever()
        
    def stop(self):
        '''
        Forcefully stops and terminates the bot's process, shutting down its connection with Discord.
        '''
        
        self.__gateway__.stop()
        self.observer.stop()

    @property 
    def eventloop(self):
        if not hasattr(self, '__eventloop__') or self.__eventloop__ is None:
            loop = asyncio.new_event_loop()
            loop.run_forever()
            self.__eventloop__ = loop
            return loop
        else:
            return self.__eventloop__
        
    async def __authenticate__(self):
        
        try:
            bot = await self.__gateway__.manager.clientInfo()
        except:
            bot = {}
            
        user = await self.__gateway__.manager.me()
        self.user = User(self, user)
        
        try:
            self.id, self.name, self.description = user.get('id'), bot.get('name'), bot.get('description')
            self.icon, self.type = bot.get('icon'), bot.get('type')
            self.summary, self.isPublic, self.requiresCodeGrant = bot.get('summary'), bot.get('bot_public', False), bot.get('bot_require_code_grant')
            self.verifyKey, self.flags, self.hook, self.isMonetized = bot.get('verify_key'), bot.get('flags'), bot.get('hook'), bot.get('is_monetized')
            self.redirectUris, self.interactionsEndpointUrl, self.roleConnectionsVerificationUrl = bot.get('redirect_uris'), bot.get('interactions_endpoint_url'), bot.get('role_connections_verification_url')
            self.owner, self.guilds, self.interactionsEventTypes = bot.get('owner'), bot.get('approximate_guild_count'), bot.get('interactions_event_types')
            self.interactionsVersion, self.explicitContentFilter, self.rpcApplicationState = bot.get('interactions_version'), bot.get('explicit_content_filter'), bot.get('rpc_application_state')
            self.storeApplicationState, self.monetizationState, self.verificationState = bot.get('store_application_state'), bot.get('creator_monetization_state'), bot.get('verification_state')
            self.public, self.discoverabilityState, self.discoveryEligibilityFlags = bot.get('integration_public'), bot.get('integration_require_code_grant'), bot.get('discoverability_state')
            self.team, self.verified, self.bio, self.locale = bot.get('team'), bot.get('verified'), user.get('bio'), user.get('locale')
            self.internalGuildRestriction, self.avatar, self.publicFlags = bot.get('internal_guild_restriction'), user.get('avatar'), user.get('public_flags')
            self.accentColor, self.banner, self.bannerColor = user.get('accent_color'), user.get('banner'), user.get('banner_color')
            self.mfa, self.mfaEnabled, self.email = user.get('mfa_enabled'), user.get('mfa_enabled'), user.get('email')
            self.displayName, self.avatarDecorationData = user.get('global_name'), user.get('avatar_decoration_data')
            self.monetizationEligibilityFlags = bot.get('monetization_eligibility_flags')
            self.username, self.discriminator = user.get('username'), user.get('discriminator')
            
            if self.displayName is not None:
                self.displayName: str = self.displayName
                self.name = self.displayName
                
            if self.id:
                self.id = int(self.id)

            self.owner: User = await self.fetchUser(
                id = self.owner.get("id")
            )
            
        except:
            import traceback; traceback.print_exc()
        
    @property
    def manager(self):
        manager: Manager = Utils.manager(self)
        return manager
    
    @property
    def tokenPrefix(self):
        '''
        The token's prefix ("Bot" or "Bearer")
        '''
        
        return 'Bot' if self.isBot else 'Bearer'
    
    @property
    def fullToken(self):
        '''
        The full token including its prefix, eg "Bearer {Token}" or "Bot {Token}"
        '''
        
        return self.tokenPrefix + ' ' + self.token
        
    def event(self, type: int = None, **filters):
        '''
        Create an event listener. The type represents the type of event to listen for. It's not required, and if not provided, the event type will be based on your listener function's name - for example, if your function's name is "on_message", it will listen for new messages.
        
        Parameters
        ----------
        type: `int`
            The event type for the listener. Examples: `Lunarcord.ON_START`, `Lunarcord.ON_MESSAGE`...

        filters: `any`
            Additional parameters used to filter the listener's usage.
        '''
        
        def inner(func):
            name = func.__name__.lower()
            
            event = __createEvent__(
                bot = self,
                callback = func,
                name = name,
                type = type
            )
            
            return event
                
        return inner
    
    def register(self, registrable: Registrable, name: str = ...):
        '''
        Registers a new, registrable object.
        
        Parameters
        ----------
        registrable: `lunarcord.Registrable`
            The item to register. This must have __register__ method.
        '''
        
        item = type(registrable)
        
        if not issubclass(item, Registrable):
            raise RegisterError('notRegistrable')
        
        if not hasattr(item, '__register__'):
            raise RegisterError('noAttribute')
            
        if registrable.registered:
            raise RegisterError('registered')
        
        registrable.__register__(self, name)
        
    def unregister(self, registrable: Registrable):
        '''
        Un-registers an already registered registrable object.
        
        Parameters
        ----------
        registrable: `lunarcord.Registrable`
            The item to register. This must have __unregister__ method.
        '''
        
        item = type(registrable)
        
        if not issubclass(item, Registrable):
            raise UnregisterError('notRegistrable')
        
        if not hasattr(item, '__unregister__'):
            raise UnregisterError('noAttribute')
            
        if registrable not in self.registered:
            raise UnregisterError('notRegistered')
        
        registrable.__unregister__(self)
        
    def reregister(self, registrable: Registrable):
        '''
        Unregisters and registers again a registered object.
        
        Parameters
        ----------
        registrable: `lunarcord.Registrable`
            The item to re-register. This must have __reregister__ method.
        '''
        
        item = type(registrable)
        
        if not issubclass(item, Registrable):
            raise ReregisterError('notRegistrable')
        
        if not hasattr(item, '__reregister__'):
            raise ReregisterError('noAttribute')
        
        '''if registrable not in self.registered:
            raise ReregisterError('notRegistered')'''
        
        registrable.__reregister__()
        
    
    def reset(self, textCommands: bool = True, slashCommands: bool = True, events: bool = True):
        
        if textCommands:
            self.textCommands = []
            
        if slashCommands:
            self.pendingSlashCommands = []
            
        if events:
            
            for container in ['onStartFunctions', 'onMessageFunctions', 'channelUpdateFunctions', 'typingStartedFunctions', 'presenceUpdateFunctions']:
                setattr(self.__gateway__, container, [])
                
                
    def addPrefix(self, *prefix: str):
        '''
        Adds one or more new `prefixes` to the bot's text command prefixes.
        If one of the prefixes already exists, the bot completely ignores it (and goes on with the next, if it exists).
        
        Parameters
        ----------
        prefix: `str`
            The prefix (or prefixes) to add to the list of all bot prefixes.
        '''
        
        prefixes = list(prefix)
        
        for prefix in prefixes:
        
            if type(prefix) is not str:
                
                try:
                    prefix = str(prefix)
                    
                except:
                    continue
                
            prefix: str = prefix.strip()
            
            if prefix in self.prefixes:
                continue
            
            self.prefixes.append(prefix)
        
        
    def removePrefix(self, *prefix: str):
        '''
        Removes one or more already existing `prefixes` from the bot's text command prefixes.
        If the prefix doesn't already exist, a `ValueError` will be raised.
        
        Parameters
        ----------
        *prefix: `str`
            The prefix(es) to remove from the list of bot prefixes.
        '''
        
        prefixes = list(prefix)
        
        for prefix in prefixes:
        
            if type(prefix) is not str:
                
                try:
                    prefix = str(prefix)
                    
                except:
                    continue
                
            prefix: str = prefix.strip()
                
            if prefix not in self.prefixes:
                raise ValueError(f'"{prefix}" not in prefixes')
                
            self.prefixes.remove(prefix)
            
    def addMentionedPrefix(self):
        
        '''
        Creates an `onMention` prefix for the bot.
        Using this will allow users to ping the bot as a prefix for text commands.
        '''
        
        self.addPrefix(self.mention)
        
    def removeMentionedPrefix(self):
        
        '''
        Deletes an already existing `onMention` prefix from this bot.
        Mention prefixes allow users to ping the bot in order to execute text commands.
        '''
        
        try:
            
            self.removePrefix(self.mention)
            
        except:
            
            raise ValueError(f'No mention prefix to remove')
            
            
    def hasPrefix(self, *prefix: str):
        '''
        Checks if the bot has any of the given prefix(es) set to it.
        
        Parameters
        ----------
        *prefix: `str`
            The prefix or prefixes to check against.
            
        Returns
        -------
        exists: `bool`
            `True` if one of the prefixes exists. `False` if none of them do.
        '''
        
        if len(prefix) == 1:
            prefixes = [prefix[0]]
            
        else:
            prefixes = list(prefix)
            
        del prefix
        
        for prefix in prefixes:
            
            if prefix in self.__gateway__.prefixes:
                return True
            
        return False
        
        
    
    def slashCommand(self, name: str = None, description: str = None, guilds: list[int] = None, cooldown: int | float = None):
        '''
        Create a bot slash command. If not given, the name will be taken from the function name.
        
        Parameters
        ----------
        name: `str`
            The slash command name which will be displayed on Discord. If None, it will be the decorated function's name.
            
        description: `str`
            The informative description for the slash command. If not given, it will be the callback function's docstring.
            
        guilds: `list`
            A list of guild (server) IDs that this command should be available in. If not given, it will be available everywhere.

        cooldown: `int`, `float`
            A command cooldown. This will be useful if you want to use `Interaction.cooldown`.
        '''
        
        if callable(name):
            func = name
            name = func.__name__
            description = func.__doc__
            
            slashCommand = __createSlashCommand__(
                bot = self,
                callback = func,
                name = name,
                description = description,
                guilds = guilds,
                cooldown = cooldown
            )
            
            return slashCommand
        
        def inner(func):
            
            slashCommand = __createSlashCommand__(
                bot = self,
                callback = func,
                name = name,
                description = description,
            )
                    
            return slashCommand
                
        return inner
    
    def textCommand(self, name: str = None, description: str = None, aliases: list[str] = [], cooldown: int | float = None):
        '''
        Create a bot text command. If not given, the name will be taken from the callback's name.
        
        Parameters
        ----------
        name: `str`
            The text command's name, which will cause it to be triggered. If None, it will be based on the callback function's name.
            
        description: `str`
            The informative description for the text command, shown in the auto-help (if it exists). If not given, it will be the callback function's docstring.
            
        aliases: `list[str]`
            Additional names that can also trigger the command's execution.

        cooldown: `int`, `float`
            A command cooldown. This will be useful if you want to use `Context.cooldown`.
        '''
            
        aliases = list(aliases)
        
        if callable(name):
            
            # When someone uses @textCommand instead of @textCommand()
            
            func = name
            name = func.__name__
            description = func.__doc__
            
            textCommand = __createTextCommand__(
                callback = func,
                bot = self,
                name = name,
                description = description,
                aliases = aliases,
                cooldown = cooldown
            )
            
            return textCommand
        
        def inner(func):
            
            textCommand = __createTextCommand__(
                callback = func,
                bot = self,
                name = name,
                description = description,
                aliases = aliases,
                cooldown = cooldown
            )
                    
            return textCommand
                
        return inner
    
    def task(self, name: str = None, autorun: bool = True):
        
        '''
        Create a new task and start its execution in another thread. If `bot` is not provided and this is never registered from a cog or with `Bot.register()`, the task will never be ran. As long as the task is registered to some bot, the given bot will take care of it immediately.
        
        Parameters
        ----------
        activity: `callable`
            Any type of callable to be ran as this Task's main activity. This is required.
            
        name: `str`
            A name for this task. If not given, it'll be generated using the activity's name.
            
        autorun: `bool`
            If this is set to `False`, the task will not be executed by the bot automatically, but by you (eg. if the activity name is `myActivity`, then `myActivity()`)
            
        bot: `Bot`
            The bot that should take care of this task and its activity. If not given, the task can't start.
        '''
        
        def create(activity):
            
            nonlocal self, name, autorun
        
            return __createTask__(
                bot = self,
                activity = activity,
                name = name,
                autorun = autorun
            )
            
        return create
    
    def loop(self, name: NameLike[str] = None, every: TimeLike[int | float] = 1, iterable: Iterable = None):
        
        '''
        Create a new loop and start it in another thread. If `bot` is not provided and this is never registered from a cog or with `Bot.register()`, the loop will never start. As long as this loop is registered to some bot, the bot will take care of it immediately.
        
        Parameters
        ----------
        name: `NameLike`
            A name for this task. If not given, it'll be generated using the activity's name.
            
        iterable: `Iterable`
            An iterable (for example a `list`) to pass items from in every iteration. This can be useful for automatically updating the bot's activity, and not only.
            
        every: `TimeLike`
            Seconds to wait before every execution of the loop's activity.
        '''
        
        def create(activity):
            
            nonlocal name, every
        
            return __createLoop__(
                bot = self,
                activity = activity,
                name = name,
                every = every,
                iterable = iterable
            )
            
        return create
        
    def __getCommandsOf(self, path: str) -> list[Registrable]:
        
        '''
        Get all attributes of type `SlashCommand`, `TextCommand` or `Event` in the module represented by the given `path`.
        
        Parameters
        ----------
        path: `str`
            The path of the target module.
            
        Returns
        -------
        commands: `list`
            A list of all `lunarcord.SlashCommand`/`lunarcord.TextCommand`/`lunarcord.Event` objects in the module.
        '''
        
        if not isfile(path):
            path += '.py'
            
            if not isfile(path):
                return []
            
        head, tail = split(path)
        initFile = join(head, '__init__.py')
        locations = []
        
        if isfile(initFile):
            hasInit = True
            locations.append(initFile)
            
        
        moduleName = split(path)[0].replace('.py', '')
        module = Utils.module(moduleName, path, locations)

        for x in ("bot", "client"):
            
            if hasattr(module, x) and getattr(module, x) is ...:
                
                setattr(module, x, self)
        
        commands = []
        
        for name in dir(module):
            item = getattr(module, name)
            
            if issubclass(type(item), Registrable):
                item.varname = name
                commands.append(item)
            
        return commands
    
    def __loadModuleFrom(self, __path: str, name = 'LunarcordModule', execute: bool = True):
        
        spec = spec_from_file_location(name, __path)
        module = module_from_spec(spec)
        
        if execute:
            spec.loader.exec_module(module)
        
        return module
    
    def __loadMainFile(self):
        
        return None
        
        time.sleep(1)
        
        module = self.__loadModuleFrom(
            self.mainfile,
            name = 'LunarcordMain'
        )
        
        regs = []
            
        return regs
            
    
    def loadFile(self, path: str, exceptions: list[str] = [], ignoreExisting: bool = False):
        
        '''
        Load all slash and text commands as well as events in the module representing the given path.
        
        You can consider this some sort of "cog".
        
        Parameters
        ----------
        path: `str`
            The path of the module containing `lunarcord.SlashCommand`, `lunarcord.TextCommand` and/or `lunarcord.Event` objects.
            
        exceptions: `list[str]`
            A list of command or event names. All registrations in this file with a name contained in this list will be ignored.
            
        Returns
        -------
        commands: `list`
            A list of all commands and events succesfully added to the bot.
        '''
        
        if path not in self.registered:
            self.registered[path] = []
        
        succesful = []
        commands = self.__getCommandsOf(path)
        namedCommands: dict[str, Registrable] = {x.__name__: x for x in self.registered[path] if hasattr(x, "__name__")}
        
        for command in commands:
                
            if ignoreExisting:

                if hasattr(command, "name") and command.name in namedCommands:

                    existing = namedCommands.get(command.__name__)
                    existingSource = Utils.functionToString(existing, imports=False)
                    source = Utils.functionToString(command, imports=False)

                    if source == existingSource:

                        continue # Continue because it already exists
            
            try:
                if ((hasattr(command, 'name'))  and(command.name in exceptions)):
                    continue # Ignore command because it is in exceptions.
                
                if True: # Treat Registrables
                    self.register(command, command.varname)
                
                self.registered[path].append(command) # Append to the list of all registrables.
                succesful.append(command) # Append to the list of succesfully added items.
                
            except:
                
                import traceback
                traceback.print_exc()
                continue # ignore errors and proceed to the next command.
        
        self.observer.addFile(path)
        return succesful # Return all succesfully added commands.
    
    def unloadFile(self, path: str) -> list[Registrable]:
        
        '''
        Un-load all previously loaded slash and text commands as well as events in the module representing the given path.
        
        Parameters
        ----------
        path: `str`
            The path of the module containing `lunarcord.SlashCommand`, `lunarcord.TextCommand` and/or `lunarcord.Event` objects.
        
        Returns
        -------
        removed: `list[Registrable]`
            All registrables (Text or Slash Commands and Events) that were previously registered from this file before the removal.
        '''
        
        if path not in self.registered:
            self.registered[path] = []
            
        before = []
        
        for regist in self.registered[path]:
            
            try:
                
                gw = self.__gateway__

                if False: # Disabled stuff
                
                    try: self.unregister(regist)
                    except: pass # This doesnt even work I think
                
                for l in self.textCommands, gw.slashCommands, gw.onStartFunctions, gw.onMessageFunctions, gw.reactionAddFunctions, gw.channelUpdateFunctions, gw.messagePinnedFunctions, gw.typingStartedFunctions, gw.typingStoppedFunctions, gw.messageDeletedFunctions, gw.messageUpdatedFunctions, gw.presenceUpdateFunctions, gw.reactionRemoveFunctions, gw.views, gw.tasks, gw.loops:
                    
                    found = False
                    
                    if regist in l:

                        found = True
                        l.remove(regist)
                        
                    if found:

                        before.append(regist)
                        
                        try: gw.views.remove(regist)
                        except: pass

                        if hasattr(regist, "__unregister__"):
                            try: regist.__unregister__()
                            except: pass
                
            except:
                ... # Ignore error and continue
        
        self.observer.removeFile(path)
        self.registered[path] = []
        return before
                
        
    
    def reloadFile(self, path: str):
        
        '''
        Re-load all previously loaded slash and text commands as well as events in the module representing the given path.
        
        Parameters
        ----------
        path: `str`
            The path of the module containing `lunarcord.SlashCommand`, `lunarcord.TextCommand` and/or `lunarcord.Event` objects.

        Returns
        -------
        commands: `list`
            A list of all commands and events that have been succesfully updated.
        '''

        from .ui import View
        
        before = self.unloadFile(path)
        after = self.loadFile(path, ignoreExisting=False)
        
        return [x for x in after if type(x) is TextCommand], [x for x in after if type(x) is SlashCommand], [x for x in after if type(x) is Event], [x for x in after if type(x) is View], [x for x in after if type(x) is Task], [x for x in after if type(x) is Loop]
    
    def unload(self) -> list[Registrable]:
        '''
        Un-load all previously loaded files and folders, removing all commands or `lunarcord.Registrable` objects of any subclassed type.
        '''
        
        unloaded = []
        
        for file in self.registered:
            
            commands = self.unloadFile(file)
            unloaded.extend(commands)
            
        return unloaded
    
    def reload(self) -> list[Registrable]:
        '''
        Re-load all previously loaded files and folders, updating and adding any new commands or `lunarcord.Registrable` objects of any subclassed type.
        '''
        
        reloaded = []
        
        for file in self.registered:
            
            files = self.reloadFile(file)
            reloaded.extend(files)
            
        return reloaded
                
    def loadFolder(self, directory: str, ignoreInit: bool = True, fileExceptions: list[str] = [], commandExceptions: list[str] = [], *, recursive: bool = True):
        
        '''
        Go through all the files in the given folder, and add their slash/text commands and events to the bot.
        
        This is almost the same as doing:
        
        ```
        for file in os.listdir(directory):
            path = os.path.join(directory, file)
            bot.loadFile(path)
        ```
        
        ... but with some extra additions to make your life easier.
        
        Parameters
        ----------
        directory: `str`
            The path of the folder to load commands, views and events from.
            
        ignoreInit: `bool`
            Whether to ignore the folder's `__init__.py` file, if it exists. By default, this is `True`.
            
        fileExceptions: `list[str]`
            A list of file paths. All files in this folder that have a name contained in this list will be ignored.
            
        commandExceptions: `list[str]`
            A list of command or event names. All commands/events which have a name contained in this list will be ignored.
            
        recursive: `bool`
            Whether to also load files of sub-folders of this folder. This is True, by default.
            
        Returns
        -------
        registrations: `list`
            A list of all commands, views and events that have been succesfully added to the bot.
        '''
        
        def checklist(item):
            if not isinstance(item, list):
                if hasattr(item, '__iter__'): # is iterable
                    item = list(item)
                else:
                    if item is None:
                        item = []
                    else:
                        item = [item]
            return item
        
        fileExceptions = checklist(fileExceptions)
        commandExceptions = checklist(commandExceptions)
                
        if ignoreInit:
            fileExceptions.append('__init__.py')
            
        fileExceptions.append('__pycache__')
        
        commands = [] # Create empty list of successes.
        files = listdir(directory) # Get all files of the directory.
        
        for file in files:
            
            path = join(directory, file) # Retrieve the full path of module.
            
            if (path in fileExceptions) or (file in fileExceptions):
                continue # Ignore the command, as expected.
            
            if os.path.isdir(path):
                
                if recursive:
                    
                    recursed = self.loadFolder(path, ignoreInit, fileExceptions, commandExceptions)
                    commands.extend(recursed) # Add all recursively loaded commands to the list
                
                continue
            
            try:
                successes = self.loadFile(path, commandExceptions) # Add and get all command in the file.
                commands.extend(successes) # Append all commands added to list of succesfully added commands.
                
            except:

                import traceback
                traceback.print_exc()
                continue # Ignore errors and continue with next file.
            
        return commands # Return all succesfully added commands.
    
    def unloadFolder(self, directory: str, ignoreInit: bool = True, fileExceptions: list[str] = [], commandExceptions: list[str] = [], *, recursive: bool = True):
        
        '''
        Go through all the files in the given folder, and remove their slash/text commands and events from the bot, if they are in.
        
        This is almost the same as doing:
        
        ```
        for file in os.listdir(directory):
            path = os.path.join(directory, file)
            bot.unloadFile(path)
        ```
        
        ... but with some extra additions to make your life easier.
        
        Parameters
        ----------
        directory: `str`
            The path of the folder to unload commands, views and events from.
            
        ignoreInit: `bool`
            Whether to ignore the folder's `__init__.py` file, if it exists. By default, this is `True`.
            
        fileExceptions: `list[str]`
            A list of file paths. All files in this folder that have a name contained in this list will be ignored.
            
        commandExceptions: `list[str]`
            A list of command or event names. All commands/events which have a name contained in this list will be ignored.
            
        recursive: `bool`
            Whether to also unload files of sub-folders of this folder. This is True, by default.
            
        Returns
        -------
        registrations: `list`
            A list of all commands, views and events that have been succesfully removed from the bot.
        '''
        
        # Time for some copypasting from loadFile
        
        def checklist(item):
            if not isinstance(item, list):
                if hasattr(item, '__iter__'): # is iterable
                    item = list(item)
                else:
                    if item is None:
                        item = []
                    else:
                        item = [item]
            return item
        
        fileExceptions = checklist(fileExceptions)
        commandExceptions = checklist(commandExceptions)
                
        if ignoreInit:
            fileExceptions.append('__init__.py')
        
        commands = [] # Create empty list of successes.
        files = listdir(directory) # Get all files of the directory.
        
        for file in files:
            
            path = join(directory, file) # Retrieve the full path of module.
            
            if (path in fileExceptions) or (file in fileExceptions):
                continue # Ignore the command, as expected.
            
            if os.path.isdir(path):
                
                if recursive:
                    
                    recursed = self.unloadFolder(path, ignoreInit, fileExceptions, commandExceptions)
                    commands.extend(recursed) # Add all recursively loaded commands to the list
                
                continue
            
            try:
                successes = self.unloadFile(path, commandExceptions) # Add and get all command in the file.
                commands.extend(successes) # Append all commands added to list of succesfully added commands.
                
            except:
                continue # Ignore errors and continue with next file.
            
        self.observer.addFile(directory)
            
        return commands # Return all succesfully removed commands.
    
    def reloadFolder(self, directory: str, ignoreInit: bool = True, fileExceptions: list[str] = [], commandExceptions: list[str] = [], *, recursive: bool = True):
        
        '''
        Go through all the files in the given folder, reloading and updating all their slash/text commands and events.
        
        This is almost the same as doing:
        
        ```
        for file in os.listdir(directory):
            path = os.path.join(directory, file)
            bot.reloadFile(path)
        ```
        
        ... but with some extra additions to make your life easier.
        
        Parameters
        ----------
        directory: `str`
            The path of the folder to load commands and events from.
            
        ignoreInit: `bool`
            Whether to ignore the folder's `__init__.py` file, if it exists. By default, this is `True`.
            
        fileExceptions: `list[str]`
            A list of file paths. All files in this folder that have a name contained in this list will be ignored.
            
        commandExceptions: `list[str]`
            A list of command or event names. All commands/events which have a name contained in this list will be ignored.
            
        recursive: `bool`
            Whether to also load files of sub-folders of this folder. This is True, by default.
            
        Returns
        -------
        registrations: `list`
            A list of all commands, views and events that have been succesfully added to the bot.
        '''
        
        def checklist(item):
            if not isinstance(item, list):
                if hasattr(item, '__iter__'): # is iterable
                    item = list(item)
                else:
                    if item is None:
                        item = []
                    else:
                        item = [item]
            return item
        
        fileExceptions = checklist(fileExceptions)
        commandExceptions = checklist(commandExceptions)
                
        if ignoreInit:
            fileExceptions.append('__init__.py')
        
        commands = [] # Create empty list of successes.
        files = listdir(directory) # Get all files of the directory.
        
        for file in files:
            path = join(directory, file) # Retrieve the full path of module.
            
            if (path in fileExceptions) or (file in fileExceptions):
                continue # Ignore the command, as expected.
            
            if os.path.isdir(path):
                
                if recursive:
                    
                    recursed = self.reloadFolder(path, ignoreInit, fileExceptions, commandExceptions)
                    commands.extend(recursed) # Add all recursively loaded commands to the list
                
                continue
            
            try:
                successes = self.reloadFile(path, commandExceptions) # Add and get all command in the file.
                commands.extend(successes) # Append all commands added to list of succesfully reloaded commands.
                
            except:
                continue # Ignore errors and continue with next file.
            
        return commands # Return all succesfully added commands.
    
    def loadFolders(self, *folders: str):

        """
        Loads all provided folders using a for loop and `Bot.loadFolder`.
        If you want more control over the way your folders are loaded (additional parameters), you should manually use `Bot.loadFolder` for each of your folders.
        Additionally, this will ignore any errors and exceptions normally raised by `loadFolder`.

        Returns
        -------
        registrations: `list`
            A list of all commands, views and events that have been succesfully added to the bot.
        """

        regist = []

        for folder in folders:

            try:
                regist.extend(
                    self.loadFolder(folder)
                )

            except:
                continue

        return regist
    
    def __fileUpdated__(self, path: str):
        
        '''
        Called when an observed file or folder is updated, from the `Observer`'s `updated` signal.
        '''
        
        try:
            text, slash, events, views, tasks, loops = self.reloadFile(path)
            text, slash, events, views, tasks, loops = len(text), len(slash), len(events), len(views), len(tasks), len(loops)

            self.__updateSlashCommandsSync__()

            items = ""
            data = []

            for x in ( ("text command", text), ("slash command", slash), ("event", events), ("view", views), ("task", tasks), ("loop", loops) ):

                name, value = x

                if value > 0:
                    s = Utils.ending(value)
                    string = f"{value} {name}{s}"
                    data.append(string)

            if data:

                useand = False
                
                if len(data) > 1:
                    remains = data.pop()
                    useand = True

                items = ", ".join(data)

                if useand:
                    items += f" and {remains}"

            if items != "":
                print(f'[-] Updated {items} in "{path}".')
            
        except:
            # print(f'[!] Failed to update commands of "{path}".')
            pass
        
    def __fileDeleted__(self, path: str):
        
        '''
        Called when an observed file or folder is deleted, from the `Observer`'s `deleted` signal.
        '''
        
        try:
            cmds = self.unloadFile(path)
            self.__updateSlashCommandsSync__()
            amt = len(cmds)
            print(f'[-] Deleted {amt} commands from missing file "{path}".')
            
        except:
            ...
    
    def setStatus(self, type: str = None, activity: Activity = None):

        from .basetypes import Status
        
        invalidType = None

        if isinstance(type, Status):

            status: Status = type
            
            if status.activity is not None:
                
                activity = status.activity

            mobile = status.mobile if isinstance(status.mobile, bool) else False
            type = status.type
        
        if not isinstance(activity, Activity) and activity is not None:
            
            if isinstance(activity, str):
                
                found = False
                
                isact = activity.lower().startswith
                stringtypes = (('playing', Playing), ('watching', Watching), ('streaming', Streaming), ('listening to', Listening), ('listeningto', Listening), ('listening', Listening))
                
                for stringtype in stringtypes:
                    stringtype, classtype = stringtype
                    
                    if isact(stringtype):
                        
                        game: str = Utils.removeFirst(stringtype, activity)
                        activity = classtype(game.strip())
                        found = True
                        break
                    
                if not found:
                    activity = Playing(activity)
                    
        if type not in ('online', 'idle', 'dnd', 'invisible', None):
            
            if not isinstance(type, Status):
                invalidType = 1
            
        if invalidType is not None:
            raise StatusError(invalidType)
        
        if activity is not None:
            activities = [activity.toJson()]
            
        else:
            if not hasattr(self, '__activity__') or self.__activity__ is None:
                activities = []
            else:
                if isinstance(self.__activity__, list) and len(self.__activity__) < 1:
                    activities = [self.__activity__]
                else:
                    activities = [self.__activity__]
                
        if type is None:
            if not hasattr(self, '__statustype__'):
                type = 'online'
                
            else:
                type = self.__statustype__ if self.__statustype__ is not None else 'online'
            
        presence = {
            'status': type,
            'afk': False,
            'activities': activities
        }
        
        coro = self.__gateway__.changeStatus(
            activities=activities,
            status=type
        )
        
        self.__gateway__.doTask(coro)
            
        try:
            self.__activity__ = activities[0]
        except:
            ... # Leave activity as is
        self.__statustype__ = type
        self.presence = presence
        
    @property
    def activity(self):
        '''
        The bot's activity. Must be a `lunarcord.Activity` (for example `lunarcord.Activity(lunarcord.PLAYING, 'with Lunarcord')` OR a `str` such as `"Playing with Lunarcord"`).
        
        When this property is changed (`bot.activity = ...`), the bot's status will instantly be updated - if it's valid.
        '''
        
        return self.__activity__
    
    @activity.setter
    def activity(self, value):
        self.setStatus(activity=value)
        
    @activity.deleter
    def activity(self):
        self.__activity__ = None
        self.setStatus(None, None)
        
    @property
    def status(self):
        '''
        The bot's status type. It can be `lunarcord.ONLINE`, `lunarcord.DND`, `LUNARCORD.IDLE`, `lunarcord.OFFLINE` or an alias of these. It can also be a `string` (such as `"online"`, `"offline"`, `"idle"`, `"dnd"`).
        
        When this property is changed (`bot.status = ...`), the bot's status will instantly be updated - if it's valid.
        '''
        
        return self.__statustype__
    
    @status.setter
    def status(self, value):
        self.setStatus(value, None)
        
    @status.deleter
    def status(self):
        self.__statustype__ = 'online'
        self.setStatus(None, None)
        
    def getUser(self, id: int) -> User:
        
        '''
        Get and return the `User` of the corresponding `id` from the cache, if found.
        
        Parameters
        ----------
        id: `int`
            The target user's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        user: `User`
            A `lunarcord.User` object representing the target user, or `None` if not found.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        for user in self.__gateway__.users:
            if user.id == id:
                return user
            
    async def fetchUser(self, id: int) -> User:
        
        '''
        Get and return the `User` of the corresponding `id` from the cache, or create it on request if not found.
        
        Parameters
        ----------
        id: `int`
            The target user's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        user: `User`
            A `lunarcord.User` object representing the target user.
        '''
        
        if type(id) is not int:
            
            try:
                id = int(id)
                
            except:
                return None
            
        found = self.getUser(id)
        
        if found is not None:
            return found
        
        try:
        
            user = await User.fromID(
                bot = self,
                user = id
            )
            
        except:
            
            user = None
            
        return user
        
    def getChannel(self, id: int) -> Channel:
        
        '''
        Get and return the `Channel` of the corresponding `id` from the cache, if found.
        
        Parameters
        ----------
        id: `int`
            The target channel's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        channel: `Channel`
            A `lunarcord.Channel` object representing the target messageable, or `None` if not found.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        for channel in self.__gateway__.channels:
            if channel.id == id:
                return channel
            
    async def fetchChannel(self, id: int) -> Channel:
        
        '''
        Get and return the `Channel` of the corresponding `id` from the cache, or create it on request if not found.
        
        Parameters
        ----------
        id: `int`
            The target channel's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        channel: `Channel`
            A `lunarcord.Channel` object representing the target messageable.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        found = self.getChannel(id)
        
        if found is not None:
            return found
        
        try:
            
            channel = await Channel.fromID(
                bot = self,
                channel = id
            )
            
        except:
            
            channel = None
        
        return channel
        
    def getMessage(self, id: int) -> Message:
        
        '''
        Get and return the `Message` of the corresponding `id` from the cache, if found.
        
        Parameters
        ----------
        id: `int`
            The target message's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        message: `Message`
            A `lunarcord.Message` object representing the target message, or `None` if not found.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        for message in self.__gateway__.messages:
            if message.id == id:
                return message
            
    async def fetchMessage(self, channel: int, id: int) -> Message:
        
        '''
        Get and return the `Message` of the corresponding `id` from the cache, or create it on request if not found.
        
        Parameters
        ----------
        channel: `int`
            The channel ID in which the message is located. Must be an integer or convertible to integer.
            
        id: `int`
            The target message's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        message: `Message`
            A `lunarcord.Message` object representing the target message.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        found = self.getMessage(id)
        
        if found is not None:
            return found
        
        try:
        
            message = await Message.fromID(
                bot = self,
                channel = channel,
                message = id
            )
            
        except:
            
            message = None
            
        return message
    
    def getGuild(self, id: int) -> Guild:
        
        '''
        Get and return the `Guild` of the corresponding `id` from the cache, if found.
        
        Parameters
        ----------
        id: `int`
            The target guild's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        guild: `Guild`
            A `lunarcord.Guild` object representing the target guild, or `None` if not found.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        for guild in self.__gateway__.guilds:
            if guild.id == id:
                return guild
            
    async def fetchGuild(self, id: int) -> Guild:
        
        '''
        Get and return the `Guild` of the corresponding `id` from the cache, or create it on request if not found.
        
        Parameters
        ----------
        id: `int`
            The target Guild's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        guild: `Guild`
            A `lunarcord.Guild` object representing the target guild/server.
        '''
        
        if type(id) is not int:
            
            try:
                id = int(id)
                
            except:
                return None
            
        found = self.getGuild(id)
        
        if found is not None:
            return found
        
        try:
        
            guild = await Guild.fromID(
                bot = self,
                guild = id
            )
            
        except:
            
            guild = None
            
        return guild
    
    def getRole(self, id: int) -> Role | None:
        
        '''
        Get and return the `Role` of the corresponding `id` from the cache, if found.
        
        Parameters
        ----------
        id: `int`
            The target role's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        role: `Role`
            A `lunarcord.Role` object representing the target role, or `None` if not found.
        '''
        
        if type(id) is not int:
            try:
                id = int(id)
                
            except:
                return None
            
        for role in self.__gateway__.roles:
            if role.id == id:
                return role
            
    def getMember(self, guild: Guild | int, id: int) -> Member | None:

        """
        Get and return the `Member` of the corresponding `id` in the given `guild` from the cache, if found.
        
        Parameters
        ----------
        guild: `Guild`, `int`
            The guild this member belongs to. Must be a guild, an integer, or a convertible to integer.
        id: `int`
            The target user's ID. Must be an integer or convertible to integer.
            
        Returns
        -------
        member: `Member`
            A `lunarcord.Member` object representing the target member, or `None` if not found.
        """

        if type(guild) is not Guild:
            guild = self.getGuild(guild)

        if guild is None:
            return
        
        try:
            id = int(id)
        except:
            return
        
        for member in guild.members:
            if member.id == id:
                return member
            
    async def fetchMember(self, guild: Guild | int, id: int) -> Member:

        """
        Get and return the `Member` of the corresponding `id` from the cache, or create it on request if not found.
        
        Parameters
        ----------
        guild: `Guild`, `int`
            The guild this member is part of. Must be a `Guild` object, an integer, or convertible to integer.

        id: `int`
            The user ID of the target member. Must be an integer or convertible to integer.
            
        Returns
        -------
        member: `Member`
            A `lunarcord.Member` object representing the requested member.
        """

        found = self.getMember(guild, id)

        if found:
            return found
        
        if type(guild) is not Guild:
            guild = await self.fetchGuild(guild)

        data: dict = await self.__gateway__.manager.loadMember(guild.id, id)
        user: dict = data.pop("user")

        return await Member.__create__(
            bot = self,
            user = user,
            data = data,
            guild = guild
        )

    
    def getChannelNamed(self, name: str) -> Channel:
        
        '''
        Get and return the `Channel` with the given `name` from the cache, if found.
        
        Parameters
        ----------
        name: `str`
            The target channel's name. Must be a string, or convertible to string.
            
        Returns
        -------
        channel: `Channel`
            A `lunarcord.Channel` object representing the target messageable, or `None` if not found.
        '''
        
        if type(id) is not str:
            try:
                id = str(id)
                
            except:
                return None
            
        for channel in self.__gateway__.channels:
            if channel.name == name:
                return channel
            
    def getUserNamed(self, name: str) -> User:
        
        '''
        Get and return the `User` with the given `name` from the cache, if found.
        This works with both display name (if existing) and real username.
        
        Parameters
        ----------
        name: `str`
            The target user's name. Must be a string, or convertible to string.
            
        Returns
        -------
        channel: `User`
            A `lunarcord.User` object representing the target user, or `None` if not found.
        '''
        
        if type(id) is not str:
            try:
                id = str(id)
                
            except:
                return None
            
        for user in self.__gateway__.users:
            if user.displayName == name or user.username == name:
                return user
            
    def getGuildNamed(self, name: str) -> Guild:
        
        '''
        Get and return the `Guild` with the given `name` from the cache, if found.
        
        Parameters
        ----------
        name: `str`
            The target guild's name. Must be a string, or convertible to string.
            
        Returns
        -------
        guild: `Guild`
            A `lunarcord.Guild` object representing the target guild, or `None` if not found.
        '''
        
        if type(id) is not str:
            try:
                id = str(id)
                
            except:
                return None
            
        for guild in self.__gateway__.guilds:
            if guild.name == name:
                return guild
            
    def getRoleNamed(self, name: str) -> Role:
        
        '''
        Get and return the `Role` with the given `name` from the cache, if found.
        
        Parameters
        ----------
        name: `str`
            The target role's name. Must be a string, or convertible to string.
            
        Returns
        -------
        role: `Role`
            A `lunarcord.Role` object representing the target role, or `None` if not found.
        '''
        
        if type(id) is not str:
            try:
                id = str(id)
                
            except:
                return None
            
        for role in self.__gateway__.roles:
            if role.name == name:
                return role
            
    def getView(self, name: str):

        """
        Gets a view with the given name. If you hadn't given your `View` a name when creating it, the variable name you used works too.

        Parameters
        ----------
        name: `str`
            The target view's custom name.

        Returns
        -------
        view: `View`, `None`
            The view that was found. Can be `None` if no view was found.
        """

        for view in self.__gateway__.views:
            if view.name == name:
                return view
    
    async def waitForMessage(
        
        self,
        author: User | Snowflake = Anyone,
        timeout: int | float = NoTimeout,
        catch: bool = False
        
    ) -> Message | None:
        
        '''
        Wait until a new message is sent by anyone if `author` is not given or is None, otherwise for a message from the specified `author`. If `timeout` passes without any messages being sent, raise an `asyncio.TimeoutError`.
        
        Parameters
        ----------
        author: `User`, `Snowflake`
            A `User` object - or an integer (or a convertible to integer) representing a Discord `Snowflake`. If this is None (or not given at all), messages by any user will be sent back.
        
        timeout: `int`, `float`
            A float or convertible to float in which time the target message should have been sent. If `None`, no error will be raised in any cases.
            
        catch: `bool`
            Whether to return `None` instead of raising an exception in case of a timeout. By default this is `False`.
        '''
        
        if author == Anyone:
            author = None
            
        if timeout == NoTimeout:
            timeout = None
            
        if author is not None:
            
            if type(author) is not User:
                
                try:
                    author = await self.fetchUser()
                    
                except:
                    author = None
            
        if type(timeout) is not float:
            
            try:
                timeout = float(timeout)
            except:
                timeout = None
                
        if type(catch) is not bool:
            
            try:
                catch = bool(catch)
                
            except:
                catch - False
                
        checked = False
        success = True
        result = None
        
        async def check(message: Message):
            
            nonlocal author, self, checked, result, success
            
            gateway = self.__gateway__
            
            if author is not None and message.author != author:
                checked = True
                success = False
            
            try:
                
                gateway.onMessageFunctions.remove(check)
                
            except:
                
                ... # Just ignore the error.
            
            result = message
            checked = True
            
            
        self.__gateway__.onMessageFunctions.append(check)
                
        while not checked:
            continue
        
        if success:
            return result
        
            
    async def __updateSlashCommands__(self, existing: list[dict] = None):
        
        '''
        Update slash commands, adding pending ones and deleting any that are missing. This is a helper method and should be used only be lunarcord itself.
        '''
        
        if existing is None:
            
            existing = await self.__gateway__.manager.slashCommands()

        existingIDs = [int(x.get("id")) for x in existing if "id" in x]
    
        createdIDs = []
        
        for pending in self.pendingSlashCommands:
            new = await pending.toJson()
            createdIDs.append(int(new.get("id")))
        
        missingIDs = [x for x in existingIDs if x not in createdIDs]
        
        if missingIDs:
            
            await self.__gateway__.manager.deleteSlashCommands(*missingIDs)
        
        self.pendingSlashCommands = []
        
    def __updateSlashCommandsSync__(self, existing: list[dict] = None):
        
        '''
        Same as using __updateSlashCommands__, but this is no longer a coroutine function, and it also happens in a separate thread - improving speed.
        '''
        
        self.__gateway__.runInThread(
            
            function = self.__updateSlashCommands__,
            existing = existing
            
        )

    async def login(self, login: str, password: str):
        
        '''
        Login using a `login` (an `email` or a `phone number`) and the corresponding account's `password`.
        This returns the account's token (a `str`) which can later be used to run this bot.
        If the credentials are invalid or something else is wrong (eg. "New login location detected..." and such errors), the returned `token` will be `None`.
        '''
        
        try:
            
            return await self.__gateway__.manager.login(login, password)
        
        except:
            
            #import traceback
            #traceback.print_exc()
            return None
        
    async def request(self, url: str, data = None):

        """
        This method exists to make your life as a coder easier.
        No more need to use additional modules for requests!
        You can use this to perform a `GET` request on any `url` with the given `data` payload.

        Parameters
        ----------
        url: `str`
            The URL to perform the request on.

        data: `Any`
            Any data to pass as the request payload.

        Returns
        -------
        returned: `bytes`, `str`, `json` or `None`
            The data returned from the other side. Can have multiple types, and could be `None` if nothing is returned.
        """

        returns: bytes = await self.manager.get(url, data, customUrl=True, returns=bytes)

        try:
            string = returns.decode()
        except:
            return returns
        
        try:
            return json.loads(string)
        except:
            return string
        
    async def cat(self) -> str | None:

        """
        If you found this method randomly, consider yourself lucky.
        This is meant to be an easter egg, but could still be useful!

        Returns
        -------
        url: `str`
            A URL leading to a random cat picture. All images provided by `thecatapi` API!
        """

        data: list[dict] = await self.request("https://api.thecatapi.com/v1/images/search")
        return data[0].get("url", None)
    
    async def europapa(self):

        """????"""

        print("[-] Joost mode activated.")
        
        gifs = [

            "https://tenor.com/view/joost-klein-joost-joost-europapa-joost-klein-europapa-joost-klein-hey-gif-8063808885285976612",
            "https://tenor.com/view/europapa-joost-klein-gif-2406371744907006571",
            "https://tenor.com/view/europapa-joost-klein-gif-15870040348330019344",
            "https://tenor.com/view/joost-joost-klein-europapa-eurovision-contest-eurovision-2024-gif-10493321365734631409",
            "https://tenor.com/view/joost-joost-klein-erik-sussy-gif-18398024777456892539",
            "https://tenor.com/view/europapa-joost-klein-gif-2444755365992642313",
            "https://tenor.com/view/joost-joost-klein-europapa-joost-wave-joost-klein-wave-gif-317417478880584456",
            "https://tenor.com/view/joost-klein-jumping-rope-eurovision-esc-gif-14556300775786763513"

        ]

        try:

            self.activity = "Listening to Europapa"

            for gif in gifs:

                try:
                    await self.owner.dm(gif)
                except:
                    await asyncio.sleep(3)

            await self.owner.dm("https://www.youtube.com/watch?v=IiHFnmI8pxg")

        except:
            ...
        
    def __repr__(self):
        
        if hasattr(self, 'name'):
            
            return f'<Bot name={self.name}>'
        
        return f'<Bot object at {hex(id(self))}>'
    
    def __str__(self):
        
        if hasattr(self, 'name'):
            
            return self.name
        
        return f'<Bot object at {hex(id(self))}>'
    
    addPrefixes = addPrefix
    removePrefixes = removePrefix
    hasPrefixes = hasPrefix